"""
Helper functions used to analyse reflectivity using various frequency estimation techniques.
"""

from enum import Enum

from numpy import cumsum, exp, linspace, log, maximum, newaxis, pi, sqrt, trapz, zeros, zeros_like


def NDFT(q, y, d=None):
    """
    Non-uniform Discrete Fourier Transform.

    There is a faster NFFT approximation but for typical reflectivity datasets the computation time is low.
    """
    if d is None:
        # if no output grid is defined, use a regular spaced grid with same number of points
        d = linspace(0, len(q) * pi / q.max(), len(q))
    FT = zeros(d.shape, dtype=complex)
    for qi, yi in zip(q, y):
        FT += yi * exp(1j * qi * d)
    return d, FT


def STFT(q, y, d=None, tau=1.0):
    """
    Short Time Frourier Transform.
    """
    return_d = False
    if d is None:
        return_d = True
        d = linspace(0, len(q) * 2.0 * pi / (q.max() - q.min()), len(q))
    FT = zeros(len(d), dtype=complex)
    for i, di in enumerate(d):
        FT[i] = trapz(y * _G(y / 2.0 / pi - di, tau) * exp(1j * di * q), x=q)
    if return_d:
        return d, FT
    else:
        return FT


def _G(t, tau):
    return exp(-pi / 2.0 * (t / tau) ** 2)


def MexicanHat(t):
    return (1.0 - t**2) * exp(-(t**2) / 2.0)


def Morlet(t, sigma=1.0):
    return pi ** (-0.25) * exp(-(t**2) / 2.0) * (exp(1j * sigma * t) - exp(-(sigma**2) / 2.0))


def CWT(q, y, wavelet=MexicanHat, d=None, N=20):
    """
    Wavelet transform using a mexican hat function.
    """
    return_d = False
    if d is None:
        return_d = True
        d = linspace(0, len(q) * 2.0 * pi / (q.max() - q.min()), len(q))

    #    b=10**linspace(log10(pi*len(d)/(d.max()-d.min())), log10(d.max()), N)
    b = linspace(5.0, d.max(), N)

    WT = zeros((len(d), len(b)), dtype=complex)
    for i, di in enumerate(d):
        Phi = (y * exp(-1j * q * di))[:, newaxis] * wavelet((q[:, newaxis] - di) / b)
        WT[i] = trapz(Phi, x=q, axis=0)
    WT /= sqrt(b)[newaxis, :]

    if return_d:
        return d, WT
    else:
        return WT


def derivative(x, y, N):
    """
    Take discrete derivative using N neighbors with N>0. For each additional
    point needed for the derivative one edge point derivative is set to 0.
    """
    dydx = zeros_like(y)
    # calculate central derivative for each point
    dx_center = x[2:] - x[:-2]
    dxmin = dx_center[dx_center > 0].min()
    dydx_center = (y[2:] - y[:-2]) / maximum(dxmin, dx_center)
    if N == 1:
        dydx[1:-1] = dydx_center
    else:
        # average N derivatives of neighboring points
        dydx_csum = cumsum(dydx_center)
        dydx[N:-N] = (dydx_csum[2 * N - 2 :] - dydx_csum[: (2 - 2 * N) or None]) / 2.0 / (N - 1)
        for i in range(2, N):
            dydx[i] = (dydx_csum[2 * i - 2] - dydx_csum[0]) / 2.0 / (i - 1)
            dydx[-i] = (dydx_csum[2 - 2 * i] - dydx_csum[-1]) / 2.0 / (i - 1)
        dydx[1] = dydx_center[0]
        dydx[-2] = dydx_center[-1]

    return dydx


def average(y, N=20):
    N = N + N % 2 - 1
    avg = y.copy()
    csum = cumsum(y)
    if N < len(avg):
        avg[N // 2 + 1 : (-N // 2) or None] = (csum[N:-1] - csum[: -N - 1]) / N
    for i in range(1, N // 2 + 1):
        avg[i] = csum[2 * i] / (2.0 * i + 1.0)
        avg[-i - 1] = (csum[-1] - csum[-2 * i - 2]) / (2.0 * i + 1.0)
    return avg


class TransformType(str, Enum):
    fourier_transform = "FT"
    # short_time_FT = 'STFT'
    mexican_hat_WT = "MH"
    morlet_WT = "MO"


def transform(
    q,
    I,
    Qc=0.0,
    trans_type: TransformType = TransformType.fourier_transform,
    avg_correction=None,
    avgN=10,
    logI=False,
    Q4=True,
    derivate=True,
    derivN=3,
    Qmin=0.0,
    Qmax=None,
    D=None,
    wavelet_scaling=0.5,
    return_corrected=False,
):
    """
    Transform a reflectivity dataset to frequency domain.
    Allows some corrections to be made before the actual transform as
    logarithmic/Q^4 compression of the dataset, taking the derivative
    and correcting q for the critical wave vector Qc.

    :param q: Q coordinate of the measurement
    :param I: Intensity of the measurement
    :keyword Qc: Critical wave vector for the correction
    :keyword trans_type: Kind of transform to be used, at the moment either 'FT','STFT', 'MH', 'MO'
                         for Fourier Transform/short-time FT, Mexican Hat WT, Morlet WT, respectively
    :keyword avg_correction: Type of average curve correction to be made, either 'subtract' or ''devide'
    :keyword log: Perform logarithmic compression
    :keyword Q4: Perform Q^4 compression
    :keyword derivate: Take the derivative of the data
    :keyword derivN: Number of points to average the derivative for smoothing
    :keyword Qmin/Qmax: Crop dataset to given Q range, this is applied after correction for Qc
    :keyword D: Optionally give an array of D values to calculate the transform for
    :keyword wavelet_offset: Value to be used to offset Q for the wavelet transform
    :keyword return_corrected: Also return the corrected Q and I values

    :returns: D, transform  or Q, I, D, transform when return_corrected is True
    """
    Q = sqrt(maximum(0.0, q**2 - Qc**2))
    if Qmax is None:
        Qmax = Q.max()
    # if no D-range is given derive it from Q
    if D is None:
        D = linspace(0.0, len(Q) * 2.0 * pi / (Q.max() - Q.min()))

    if Q4:
        I = Q**4 * I
    if logI:
        I = log(maximum(I, I[I > 0].min()))

    if avg_correction:
        avg = average(I, N=avgN)
        if avg_correction == "subtract":
            I -= avg
        else:
            I /= avg

    if derivate:
        Drange = Q > 0
        I = I[Drange]
        Q = Q[Drange]
        I = derivative(Q, I, derivN)

    posvals = (Q > Qmin) & (Q <= Qmax)
    Quse = Q[posvals]
    Iuse = I[posvals]

    if trans_type == TransformType.fourier_transform:
        D, T = NDFT(Quse, Iuse, d=D)
        T = abs(T)
    # elif trans_type==TransformType.short_time_FT:
    #     T = abs(STFT(Quse, Iuse, D))
    elif trans_type == TransformType.mexican_hat_WT:
        T = abs(CWT(Quse, Iuse, d=D, wavelet=MexicanHat)) ** 2.0 + 1e-20
        T = T[:, int(T.shape[1] * wavelet_scaling)]
    elif trans_type == TransformType.morlet_WT:
        T = abs(CWT(Quse, Iuse, d=D, wavelet=Morlet)) ** 2.0 + 1e-20
        T = T[:, int(T.shape[1] * wavelet_scaling)]
    else:
        raise ValueError("Transformation type not known: %s" % trans_type)

    if return_corrected:
        return Q, I, D, T
    else:
        return D, T



"""
Executable module to start a server that can be used for
remote refinement on a server.
"""

import argparse
import asyncio
import logging
import os

import platformdirs

from . import version
from .core import custom_logging

config_path = os.path.abspath(platformdirs.user_data_dir("GenX3", "ArturGlavic"))
os.environ["NUMBA_CACHE_DIR"] = os.path.join(config_path, "numba_cache")


def set_numba_single():
    logging.debug("Setting numba JIT compilation to single CPU")
    os.environ["NUMBA_CACHE_DIR"] = os.path.join(config_path, "single_cpu_numba_cache")

    logging.debug("Setting numba JIT compilation to single CPU")
    import numba

    old_jit = numba.jit

    def jit(*args, **opts):
        opts["parallel"] = False
        return old_jit(*args, **opts)

    numba.jit = jit
    numba.GENX_OVERWRITE_SINGLE = True
    try:
        numba.set_num_threads(1)
    except AttributeError:
        pass


def main():
    import multiprocessing

    multiprocessing.set_start_method("spawn")
    try:
        from mpi4py import MPI
    except ImportError:
        rank = 0
        __mpi__ = False
    else:
        __mpi__ = bool(MPI.COMM_WORLD.Get_size() > 1)
        rank = MPI.COMM_WORLD.Get_rank()

    parser = argparse.ArgumentParser(
        description="GenX %s, fits data to a model. Server script for remote fitting." % version.__version__,
        epilog="For support, manuals and bug reporting see http://genx.sf.net",
    )
    parser.add_argument("address", nargs="?", default="localhost", help="Network address to listen on")
    parser.add_argument("port", nargs="?", default=3000, help="Network port to listen on")
    parser.add_argument(
        "-l", "--logfile", dest="logfile", default=None, type=str, help="Output debug information to logfile."
    )
    parser.add_argument(
        "-p",
        "--password",
        dest="password",
        default="empty",
        type=str,
        help="Password used when establishing connection from client.",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        default=False,
        action="store_true",
        help="Show additional debug information on console/logfile",
    )
    parser.add_argument(
        "--disable-nb",
        dest="disable_numba",
        default=False,
        action="store_true",
        help="Disable the use of numba JIT compiler",
    )
    parser.add_argument(
        "--no-nb1",
        dest="numba_single",
        default=True,
        action="store_false",
        help="Compile numba JIT functions with parallel computing support (use more then one core). "
        "This is the default for no-server based fits.",
    )

    args = parser.parse_args()

    if not args.debug:
        custom_logging.CONSOLE_LEVEL = logging.INFO
    if rank > 0:
        custom_logging.CONSOLE_LEVEL = logging.WARNING
    custom_logging.setup_system()

    if args.logfile:
        if __mpi__:
            lfbase, lfend = args.logfile.rsplit(".", 1)
            custom_logging.activate_logging(f"{lfbase}_{rank:02}.{lfend}")
        else:
            custom_logging.activate_logging(args.logfile)
    if rank == 0:
        logging.debug("Arguments from parser: %s" % args)

    if args.disable_numba:
        logging.debug("Manually disable numba JIT compilation")
        from genx.models import lib as modellib

        modellib.USE_NUMBA = False
    else:
        if args.numba_single:
            try:
                set_numba_single()
            except ImportError:
                pass
        if rank == 0:
            try:
                import numba
            except ModuleNotFoundError:
                logging.debug("Numba not found, don't import JIT functions.")
            else:
                logging.info("Importing numba based modules to pre-compile JIT functions, this can take some time")
                from genx.models.lib import instrument_numba, neutron_numba, offspec, paratt_numba, surface_scattering

        if rank == 0:
            logging.info("Modules imported successfully")

    if rank == 0:
        logging.info("Starting RemoteController")
    from .remote import controller

    ctrl = controller.RemoteController()
    ctrl.key = args.password.encode("utf-8")
    asyncio.run(ctrl.serve(args.address, args.port))


if __name__ == "__main__":
    main()



"""data_loader_wx.py

Implements the GUI side of data_loader as DataLoaderController which
takes care of the use of the DataLoaders.
"""

import io
import os
import traceback

import wx

from genx.core.config import config
from genx.core.custom_logging import iprint

from .utils import PluginHandler, ShowErrorDialog

head, tail = os.path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled...
__FILENAME__ = tail.split(".")[0]
# This assumes that plugin is under the current dir may need
# changing
__MODULE_DIR__ = head
if __MODULE_DIR__ != "/":
    __MODULE_DIR__ += "/"


class PluginController:
    def __init__(self, parent):
        self.plugin_handler = PluginHandler(parent, __MODULE_DIR__, "data_loaders")
        self.parent = parent
        self.plugin_handler.load_plugin("auto")

    def load_default(self):
        try:
            plugin_name = config.get("data handling", "data loader")
            self.LoadPlugin(plugin_name)
        except Exception as S:
            iprint("Could not locate the data loader parameter or the data loader. Error:")
            iprint(S.__str__())
            iprint("Proceeding with laoding the default data loader.")
            self.LoadPlugin("default")

    def LoadPlugin(self, plugin):
        """
        Loads a data handler note that there is no UnLoad function
        since only one DataHandler can be plugged in at a time.
        """
        # Unload the plugins
        names = self.plugin_handler.loaded_plugins.copy()
        try:
            [self.plugin_handler.unload_plugin(pl) for pl in names]
            self.parent.SetStatusText("Unloaded data loader %s" % list(names.keys())[0])
        except:
            outp = io.StringIO()
            traceback.print_exc(200, outp)
            tbtext = outp.getvalue()
            outp.close()
            ShowErrorDialog(
                self.parent,
                "Can NOT unload plugin object" + list(names.keys())[0] + "\nPython traceback below:\n\n" + tbtext,
            )
        try:
            self.plugin_handler.load_plugin(plugin)
            self.parent.SetStatusText("Loaded data loader: %s" % plugin)
        except:
            outp = io.StringIO()
            traceback.print_exc(200, outp)
            tbtext = outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, "Can NOT load plugin " + plugin + "\nPython traceback below:\n\n" + tbtext)

    def ShowDialog(self):
        """ShowDialog(self) --> None

        Shows a dialog boc for the user to choose a data loader.
        """
        cur_plugin = list(self.plugin_handler.loaded_plugins.keys())[0]
        plugin_list = self.plugin_handler.get_possible_plugins()
        dlg = PluginDialog(self.parent, plugin_list, cur_plugin, self.LoadPlugin)
        dlg.ShowModal()
        dlg.Destroy()


class PluginDialog(wx.Dialog):
    def __init__(self, parent, plugin_list, current_plugin, load_plugin_func=None):
        wx.Dialog.__init__(self, parent, -1, "Choose a data loader")

        self.load_plugin_func = load_plugin_func

        choice_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer = wx.BoxSizer(wx.VERTICAL)

        choice_text = wx.StaticText(self, -1, "Data loaders: ")
        self.choice_control = wx.Choice(self, -1, choices=plugin_list)
        self.choice_control.SetStringSelection(current_plugin)
        choice_sizer.Add(choice_text, 0, wx.ALIGN_CENTER_VERTICAL)
        choice_sizer.Add(self.choice_control, 0, wx.ALIGN_CENTER_VERTICAL)

        # Add the Dialog buttons
        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        apply_button = wx.Button(self, wx.ID_APPLY)
        apply_button.SetDefault()
        button_sizer.AddButton(apply_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()
        # Add some eventhandlers
        self.Bind(wx.EVT_BUTTON, self.on_apply, okay_button)
        self.Bind(wx.EVT_BUTTON, self.on_apply, apply_button)

        sizer.Add((20, 20), 0)
        sizer.Add(choice_sizer, 0, wx.ALIGN_CENTRE | wx.ALL)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.TOP, 20)
        sizer.Add(button_sizer)

        self.SetSizer(sizer)

        sizer.Fit(self)
        self.Layout()

    def on_apply(self, event):
        if self.load_plugin_func is not None:
            self.load_plugin_func(self.choice_control.GetStringSelection())
        event.Skip()



__author__ = "GenX"

import string

from functools import reduce

# noinspection PyUnresolvedReferences
import vtk.util.colors as vtkc
import wx

from numpy import arctan2, pi, sign, sqrt
# inspection does not work as vtk mangles with the python miport system
# noinspection PyUnresolvedReferences
from vtk import (vtkActor, vtkActor2D, vtkArrowSource, vtkAxesActor, vtkPolyDataMapper, vtkRenderer, vtkRenderWindow,
                 vtkRenderWindowInteractor, vtkSphereSource, vtkTextMapper, vtkTransform)

from genx.core.custom_logging import iprint

from . import atom_colors as atom_colors
from . import custom_dialog, sxrd_images
from .wxVTKRenderWindow import wxVTKRenderWindow


class VTKview(wxVTKRenderWindow):
    def __init__(self, parent):
        wxVTKRenderWindow.__init__(self, parent, wx.NewId(), stereo=0)
        self.ren = vtkRenderer()
        self.GetRenderWindow().AddRenderer(self.ren)
        self.ren.SetBackground(0, 0, 0)
        self.sphereActor = vtkActor()

        # Settings for creation of sample
        self.x_uc = 1
        self.y_uc = 1
        self.use_sym = True
        self.fold_sym = True
        self.show_bulk = True
        self.arrows_for_atoms = 0.0  # if >0, add arrows for dx,dy,dz

        # Some defualts
        self.radius = 1.0
        self.theta_res = 15
        self.phi_res = 15
        self.amb_col = (1.0, 1.0, 1.0)
        self.diffuse = 1.0
        self.ambient = 0.2
        self.specular = 0.5
        self.specular_power = 50.0
        self.specular_col = (1.0, 1.0, 1.0)
        self.element_col = atom_colors.jmol
        self.default_col = vtkc.yellow

        self.search_radius = 0.1

        self.toolbar = None
        self.cursor_mode = "orbit"

        self.textMapper = vtkTextMapper()
        tprop = self.textMapper.GetTextProperty()
        tprop.SetFontFamilyToArial()
        tprop.SetFontSize(10)
        tprop.BoldOn()
        tprop.ShadowOn()
        tprop.SetColor(1, 1, 1)
        self.textActor = vtkActor2D()
        self.textActor.VisibilityOff()
        self.textActor.SetMapper(self.textMapper)

        self.ren.AddActor(self.textActor)

    def do_toolbar(self, parent):
        """Create and return a toolbar that can be used to control the widget"""
        toolbar = wx.ToolBar(parent, style=wx.TB_FLAT | wx.TB_VERTICAL)
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        tb_bmp_size = int(dpi_scale_factor * 20)

        button_names = ["View X", "View Y", "View Z", "Isometric"]
        button_images = [sxrd_images.x, sxrd_images.y, sxrd_images.z, sxrd_images.isometric]
        callbacks = [self.OnViewX, self.OnViewY, self.OnViewZ, self.OnViewIsometric]
        tooltips = ["View along X", "View along Y", "View along Z", "Isometric view"]

        for i in range(len(button_names)):
            new_id = wx.NewId()
            toolbar.AddTool(
                new_id,
                label=button_names[i],
                bitmap=wx.Bitmap(button_images[i].GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                shortHelp=tooltips[i],
            )
            parent.Bind(wx.EVT_TOOL, callbacks[i], id=new_id)

        button_names = ["Select", "Orbit", "Zoom", "Pan"]
        button_images = [sxrd_images.selection, sxrd_images.orbit, sxrd_images.zoom_small, sxrd_images.pan]
        callbacks = [
            self.OnChangeCursorState,
            self.OnChangeCursorState,
            self.OnChangeCursorState,
            self.OnChangeCursorState,
        ]
        tooltips = ["Object selection", "Orbit", "Zoom", "Pan"]

        self.cursor_ids = []
        for i in range(len(button_names)):
            new_id = wx.NewId()
            self.cursor_ids.append(new_id)
            toolbar.AddCheckTool(
                new_id,
                button_names[i],
                wx.Bitmap(button_images[i].GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                shortHelp=tooltips[i],
            )
            parent.Bind(wx.EVT_TOOL, callbacks[i], id=new_id)

        toolbar.ToggleTool(self.cursor_ids[1], True)
        self.toolbar = toolbar

        return toolbar

    def OnViewX(self, event):
        """Aligns the view with the x axis"""
        if self._CurrentRenderer:
            self._CurrentCamera.SetFocalPoint(0, 0, 0)
            self._CurrentCamera.SetPosition(1.0, 0.0, 0.0)
            self._CurrentCamera.SetViewUp(0.0, 0.0, 1.0)

            # self._CurrentCamera.OrthogonalizeViewUp()
            self._CurrentRenderer.ResetCamera()
            self.Render()

    def OnViewY(self, event):
        """Aligns the view with the y axis"""
        if self._CurrentRenderer:
            self._CurrentCamera.SetFocalPoint(0, 0, 0)
            self._CurrentCamera.SetPosition(0.0, 1.0, 0.0)
            self._CurrentCamera.SetViewUp(0.0, 0.0, 1.0)
            self._CurrentRenderer.ResetCamera()
            self.Render()

    def OnViewZ(self, event):
        """Alings the view with the z axis"""
        if self._CurrentRenderer:
            self._CurrentCamera.SetFocalPoint(0, 0, 0)
            self._CurrentCamera.SetPosition(0.0, 0.0, 1.0)
            self._CurrentCamera.SetViewUp(0.0, 1.0, 0.0)
            self._CurrentRenderer.ResetCamera()
            self.Render()

    def OnViewIsometric(self, event):
        """Creates an isometric view"""
        if self._CurrentRenderer:
            self._CurrentCamera.SetFocalPoint(0, 0, 0)
            self._CurrentCamera.SetPosition(1.0, 1.0, 1.0)
            self._CurrentCamera.SetViewUp(0.0, 0.0, 1.0)
            self._CurrentRenderer.ResetCamera()
            self.Render()

    def OnChangeCursorState(self, event):
        """Callback when changing the cursor state between select, orbit, zoom and pan"""
        if self.toolbar:
            [self.toolbar.ToggleTool(cid, False) for cid in self.cursor_ids]
            self.toolbar.ToggleTool(event.GetId(), True)
            name = self.toolbar.FindById(event.GetId()).GetLabel()
            if self.cursor_mode == "select":
                self.textActor.VisibilityOff()
                self.sphereActor.VisibilityOff()
                self.Render()
            if name == "Select":
                self.cursor_mode = "select"
            elif name == "Orbit":
                self.cursor_mode = "orbit"
            elif name == "Zoom":
                self.cursor_mode = "zoom"
            elif name == "Pan":
                self.cursor_mode = "pan"
            else:
                iprint("VTKView.OnChangeCursorState: Button name ", name, "is not a known button")

    def highlight(self, actor):
        # outline = vtk.vtkOutlineFilter()

        sphere = vtkSphereSource()
        sphere.SetRadius(self.radius * 1.1)
        sphere.SetCenter(actor.GetCenter())
        sphere.SetThetaResolution(self.theta_res)
        sphere.SetPhiResolution(self.phi_res)

        spheremapper = vtkPolyDataMapper()
        spheremapper.SetInputConnection(sphere.GetOutputPort())

        sphereActor = vtkActor()
        sphereActor.SetMapper(spheremapper)
        sphereActor.GetProperty().SetColor((0.0, 0.0, 0.0))
        sphereActor.GetProperty().SetOpacity(0.5)

        # sphereActor.GetProperty().SetRepresentationToWireframe()
        try:
            self.ren.RemoveViewProp(self.sphereActor)
        except Exception as e:
            iprint(e)
        self.ren.AddActor(sphereActor)
        self.sphereActor = sphereActor
        self.Render()

    def OnLeftDown(self, event):
        """Overriding the defualt on left down"""
        if self.cursor_mode == "zoom":
            self._Mode = "Zoom"
        elif self.cursor_mode == "pan":
            self._Mode = "Pan"
        elif self.cursor_mode == "orbit":
            self._Mode = "Rotate"
        elif self.cursor_mode == "select":
            self.OnSelectAtom(event)
            event.Skip()

    def OnSelectAtom(self, event):
        """Callback for  the selection of an atom"""
        if self._CurrentRenderer:
            x = event.GetX()
            y = event.GetY()

            renderer = self._CurrentRenderer
            picker = self._Picker

            windowX, windowY = self._RenderWindow.GetSize()

            picker.Pick(x, (windowY - y - 1), 0.0, renderer)
            actor = picker.GetActor()
            if not picker.GetActor():
                self.textActor.VisibilityOff()
                self.sphereActor.VisibilityOff()
                self.Render()
            else:
                selPt = picker.GetSelectionPoint()
                pickPos = picker.GetPickPosition()

                x, y, z = actor.GetCenter()
                self.textMapper.SetInput("%s" % reduce(lambda x, y: x + "," + y, self.locate_ids(x, y, z)))
                self.textActor.SetPosition(picker.GetSelectionPoint()[:2])
                self.textActor.VisibilityOn()
                self.highlight(actor)

    def OnRightDown(self, event):
        """Overriding defualt Pick Actor"""
        if not self._Mode:
            if event.ControlDown():
                self._Mode = "Rotate"
            elif event.ShiftDown():
                self._Mode = "Pan"
            else:
                self._Mode = "Zoom"

    def locate_ids(self, x, y, z):
        """Given the picked positions x,y,z locate the strings ids that
        are within a distance of self.search_radius.
        """
        within_dist = ((self.atom_x - x) ** 2 + (self.atom_y - y) ** 2 + (self.atom_z - z) ** 2) < self.search_radius**2
        ids = []
        for i, pick in enumerate(within_dist):
            if pick and not self.atom_ids[i] in ids:
                ids.append(self.atom_ids[i])
        return ids

    def _get_col(self, element):
        """Translate element to color according to element dict."""
        # We need to take care if the eleemnt is an ion.
        # the two last chars are on the form Xp or Xm where X is a digit
        element = element.lower()
        if len(element) > 2:
            if element[-2] in string.digits:
                element = element[:-2]
        if element in self.element_col:
            col = self.element_col[element]
        else:
            col = self.default_col
        return col

    def clear_view(self):
        """remove all actors so a new sample can be built."""
        self.ren.RemoveAllViewProps()
        self.ren.AddActor2D(self.textActor)
        self.textActor.VisibilityOff()

    def create_axes(self, a=2.0, b=2.0, c=2.0):
        axesActor = vtkAxesActor()
        axesActor.SetShaftTypeToCylinder()
        # axesActor.SetXAxisLabelText('x')
        # axesActor.SetYAxisLabelText('y')
        # axesActor.SetZAxisLabelText('z')
        axesActor.SetTotalLength(a, b, c)
        self.ren.AddActor(axesActor)

    def _create_sphere(self, x, y, z, col, opacity=1.0):
        """Create a sphere actor at (x,y,z) with color c and return it"""
        sphere = vtkSphereSource()
        sphere.SetRadius(self.radius)
        sphere.SetCenter(x, y, z)
        sphere.SetThetaResolution(self.theta_res)
        sphere.SetPhiResolution(self.phi_res)

        spheremapper = vtkPolyDataMapper()
        spheremapper.SetInputConnection(sphere.GetOutputPort())

        sphereActor = vtkActor()
        sphereActor.SetMapper(spheremapper)
        sphereActor.GetProperty().SetDiffuseColor(col)
        sphereActor.GetProperty().SetAmbientColor(self.amb_col)
        sphereActor.GetProperty().SetDiffuse(self.diffuse)
        sphereActor.GetProperty().SetAmbient(self.ambient)
        sphereActor.GetProperty().SetSpecular(self.specular)
        sphereActor.GetProperty().SetSpecularPower(self.specular_power)
        sphereActor.GetProperty().SetSpecularColor(self.specular_col)
        sphereActor.GetProperty().SetOpacity(opacity)
        return sphereActor

    def add_atom(self, x, y, z, id, element, opacity=1.0):
        col = self._get_col(element)
        sphere_actor = self._create_sphere(x, y, z, col, opacity)
        self.ren.AddActor(sphere_actor)

    def add_atom_shifts(self, x, y, z, dx, dy, dz, element):
        col = self._get_col(element)
        # scale shift to make it more visible
        dx *= self.arrows_for_atoms
        dy *= self.arrows_for_atoms
        dz *= self.arrows_for_atoms
        length = sqrt(dx**2 + dy**2 + dz**2)
        if length == 0:
            return

        arrowSource = vtkArrowSource()
        arrowSource.SetShaftRadius(0.15 / length)
        arrowSource.SetTipRadius(0.3 / length)
        arrowSource.SetShaftResolution(16)
        arrowSource.SetTipResolution(16)
        # arrowSource.SetTipLength(0.2/length)

        # Create a mapper and actor
        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(arrowSource.GetOutputPort())
        actor = vtkActor()
        actor.SetMapper(mapper)

        transform = vtkTransform()
        transform.Translate(x, y, z)
        transform.RotateY(180.0 / pi * arctan2(dz, sqrt(dx**2 + dy**2)))
        transform.RotateZ(180.0 / pi * arctan2(dy, dx))
        transform.Scale(length, length, length)
        actor.SetUserTransform(transform)
        actor.GetProperty().SetColor(col)

        self.ren.AddActor(actor)
        return

        arrowActor = vtkAxesActor()
        arrowActor.SetShaftTypeToCylinder()
        arrowActor.AxisLabelsOff()

        arrowActor.GetXAxisCaptionActor2D().GetCaptionTextProperty().SetColor(*col)
        arrowActor.GetYAxisCaptionActor2D().GetCaptionTextProperty().SetColor(*col)
        arrowActor.GetZAxisCaptionActor2D().GetCaptionTextProperty().SetColor(*col)

        arrowActor.SetTotalLength(abs(dx), abs(dy), abs(dz))

        transform = vtkTransform()
        transform.Translate(x, y, z)
        transform.Scale(sign(dx), sign(dy), sign(dz))
        arrowActor.SetUserTransform(transform)
        self.ren.AddActor(arrowActor)

    def build_sample(self, sample, use_opacity=False):
        """Builds an sample view from the object sample that has to
        have the method _surf_pars that yields the x, y, z, u, oc, el
        """
        self.clear_view()
        x, y, z, u, oc, el, ids = sample.create_uc_output(
            x_uc=self.x_uc, y_uc=self.y_uc, use_sym=self.use_sym, fold_sym=self.fold_sym, use_bulk=self.show_bulk
        )
        self.atom_x = x
        self.atom_y = y
        self.atom_z = z
        self.atom_ids = ids
        if use_opacity:
            [self.add_atom(x[i], y[i], z[i], "None", el[i], oc[i]) for i in range(len(x))]
        else:
            [self.add_atom(x[i], y[i], z[i], "None", el[i], 1.0) for i in range(len(x))]
        if self.arrows_for_atoms > 0:
            dx, dy, dz = sample.create_shift_output(
                x_uc=self.x_uc, y_uc=self.y_uc, use_sym=self.use_sym, fold_sym=self.fold_sym, use_bulk=self.show_bulk
            )
            [self.add_atom_shifts(x[i], y[i], z[i], dx[i], dy[i], dz[i], el[i]) for i in range(len(x))]
        self.create_axes()

        self.ren.ResetCamera()
        self.Render()

    def Show(self, *args):
        self.Render()

    def show(self):
        self.Reset()
        self.Render()

    def ShowSettingDialog(self):
        """Shows a settings dialog to change the settings"""
        parameters = [
            "Use symmetry",
            "Fold symmetry op",
            "Show Bulk",
            "a unit cell rep.",
            "b unit cell rep.",
            "Atom radius",
            "Delta Arrow Scale",
        ]
        values = {
            "Use symmetry": self.use_sym,
            "Fold symmetry op": self.fold_sym,
            "Show Bulk": self.show_bulk,
            "a unit cell rep.": self.x_uc,
            "b unit cell rep.": self.y_uc,
            "Atom radius": self.radius,
            "Delta Arrow Scale": self.arrows_for_atoms,
        }
        units = {
            "Use symmetry": "",
            "Fold symmetry op": "",
            "Show Bulk": "",
            "a unit cell rep.": "uc",
            "b unit cell rep.": "uc",
            "Atom radius": "AA",
            "Delta Arrow Scale": "0=off",
        }
        validators = {
            "Use symmetry": True,
            "Fold symmetry op": True,
            "Show Bulk": True,
            "a unit cell rep.": 1,
            "b unit cell rep.": 1,
            "Atom radius": custom_dialog.FloatObjectValidator(),
            "Delta Arrow Scale": custom_dialog.FloatObjectValidator(),
        }
        groups = [
            ["Unit cell", ("Use symmetry", "Fold symmetry op", "Show Bulk", "a unit cell rep.", "b unit cell rep.")],
            ["Rendering", ("Atom radius", "Delta Arrow Scale")],
        ]

        dlg = custom_dialog.ValidateDialog(
            self, parameters, values, validators, units=units, groups=groups, title="Domain Viewer Settings"
        )
        if dlg.ShowModal() == wx.ID_OK:
            new_values = dlg.GetValues()
            self.use_sym = bool(new_values["Use symmetry"])
            self.fold_sym = bool(new_values["Fold symmetry op"])
            self.show_bulk = bool(new_values["Show Bulk"])
            self.x_uc = int(new_values["a unit cell rep."])
            self.y_uc = int(new_values["b unit cell rep."])
            self.radius = float(new_values["Atom radius"])
            self.arrows_for_atoms = float(new_values["Delta Arrow Scale"])



"""
Library for combined x-ray and neutrons simulations for inhomogeneous samples.
==============================================================================
In addition to the options from spec_nx this allows to model thickness variation over
the sample surface and gradients in the roughness and thickness over a repetition of layers.

Classes
-------
"""

from copy import deepcopy
from dataclasses import dataclass, field, fields
from typing import List

import numpy as np

from . import spec_nx
# import all special footprint functions
from .lib import footprint as footprint_module
from .lib import neutron_refl as MatrixNeutron
from .lib import paratt as Paratt
from .lib import refl_base as refl
from .lib import resolution as resolution_module
from .lib.base import AltStrEnum
from .lib.footprint import *
from .lib.instrument import *
from .lib.physical_constants import T_to_SL, muB_to_SL, r_e
from .lib.resolution import *
from .lib.sm_hayter_mook import sm_layers
from .lib.testing import ModelTestCase
from .spec_nx import Coords, FootType, Instrument
from .spec_nx import Layer as NXLayer
from .spec_nx import LayerParameters as NXLayerParameters
from .spec_nx import Polarization, Probe, ResType, q_limit

# Preamble to define the parameters needed for the models outlined below:
ModelID = "SpecInhom"

Buffer = spec_nx.Buffer

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"

@dataclass
class Layer(NXLayer):
    """
    Representing a layer in the sample structur.

    ``b``
       The neutron scattering length per formula unit in fm (fermi meter =
       1e-15m)
    ``d``
       The thickness of the layer in AA (Angstroms = 1e-10m)
    ``f``
       The x-ray scattering length per formula unit in electrons. To be
       strict it is the number of Thompson scattering lengths for each
       formula unit.
    ``dens``
       The density of formula units in units per Angstroms. Note the units!
    ``magn_ang``
       The angle of the magnetic moment in degress. 0 degrees correspond to
       a moment collinear with the neutron spin.
    ``magn``
       The magnetic moment per formula unit (same formula unit as b and dens
       refer to)
    ``sigma``
       The root mean square roughness of the top interface of the layer in
       Angstroms.
    ``xs_ai``
       The sum of the absorption cross section and the incoherent scattering
       cross section in barns for neutrons
    ``sigma_gradient``
       Increase of roughness of this layer from bottom to top in stack repetitions
    ``d_gradient``
       Increase of the thickness of this layer from bottom to top in stack repetitions
    """

    sigma_gradient: float = 0.0
    d_gradient: float = 0.0

    Units = NXLayer.Units | {"sigma_gradient": "1", "d_gradient": "1"}

    Groups = [
        ("General", ["d", "dens", "sigma"]),
        ("Neutron", ["b", "xs_ai", "magn", "magn_ang"]),
        ("Inhom.", ["sigma_gradient", "d_gradient"]),
        ("X-Ray", ["f"]),
    ]


@dataclass
class LayerParameters(NXLayerParameters):
    sigma_gradient: List[float]
    d_gradient: List[float]


class SigmaGradientType(AltStrEnum):
    rel_lin = "relative linear"
    rel_sqrt = "relative sqrt"
    abs_lin = "absolute linear"
    abs_sqrt = "absolute sqrt"


@dataclass
class Stack(refl.StackBase):
    """
    A collection of Layer objects that can be repeated.

    ``Layers``
       A ``list`` consiting of ``Layer``s in the stack the first item is
       the layer closest to the bottom
    ``Repetitions``
       The number of repsetions of the stack or m-value for analytical super-mirror model.
    ``sigma_gtype``
       model for this increase in roughness (relative linlinea, relative sqrt, absolute linear, absolute sqrt)
    ``sigma_gradient``
       amount of increase in roughness from bottom to top applied to all layers in the stack
    ``d_gradient``
       thickness increase from bottom to top
    ``dens_gradient``
       density increase from bottom to top
    ``beta_sm``
       Parameter used to model neutron super-mirror coatings. There are two models implemented.
       If *beta_sm* is positive, a simplified analystical model from J. Schelten and K. Mika (Nuc. Inst. Metho. 160 (1979))
       is used and the parameter represents the quality paramter for super-mirror sequences.
       A super-mirror sequence is generated and Repetitions is interpreted as the m-value for
       the sequence to calculate actual repetitions automatically.

       If *beta_sm* is negative it is interpretated as the zeta parameter of an iterative build-up of super-mirror
       layers using the method described by J.B. Hayter and H.A. Mook (J. Appl. Cryst. (1989), 22, 35-41). Here,
       the actual given number of repetitions are used to build up a super-mirror from the critical edge
       upowards. The m-value follows from the number of repetitions, material parameters and zeta.

       This method only works acuratly with two layers in the Stack, as the maximum and
       minimum SLD layers are used for the SLD calculation needed to get the layer sequences.
       (Other layers are simulated but their SLDs are ignored for the
       number of repetitions and super-lattice repetition period).

       For the Schelten/Mika method, all layers are scaled to the relative size needed to produce the
       right super-lattice period. In Hayter/Mook only the two min/max components are set to the
       optimal thickness values while the rest of the stack stays with constant width.

       The value of d_gradient is interpreted as a relative change from bottom to top in this case. Together
       with the *sm_scale*, that is applied to all layers.
    ``sm_scale``
       A scaling parameter applied to all thicknesses in a supermirror. This can account for thickness
       differences introduced by the manufacturing process due to imperfect calibration.
    """

    Layers: List[Layer] = field(default_factory=list)
    Repetitions: float = 1.0
    sigma_gtype: SigmaGradientType = 0  # model for this increas (rel lin, rel sqrt, abs lin, abs sqrt)
    sigma_gradient: float = 0.0  # amount of increase in roughness from bottom to top
    d_gradient: float = 0.0  # change in thickness bottom to top
    dens_gradient: float = 0.0  # change in density bottom to top
    beta_sm: float = 0.0  # function applied to change thickness (linear,  super-mirror (d=beta SM))
    sm_scale: float = 1.0  # scaling factor for thicknesses in super-mirror

    Groups = [
        ("General", ["Repetitions"]),
        ("Roughness Gradient", ["sigma_gtype", "sigma_gradient"]),
        ("Supermirror", ["beta_sm", "sm_scale"]),
        ("Other Gradients", ["d_gradient", "dens_gradient"]),
    ]

    def resolveLayerParameter(self, name):
        """
        Add gradient for sigma and thickness to multilayers
        """
        if self.beta_sm > 0:
            # user has selected super-mirror coating sequence
            m = self.Repetitions
            beta = self.beta_sm
            rho_all = [li.dens * (li.b.real + li.magn) * 1e-5 for li in self.Layers]
            rhoA = max(rho_all)
            rhoB = min(rho_all)
            # critical wave vector values (in paper this includes an addition factor of 1/2)
            QcA = 4.0 * pi * sqrt(rhoA / pi + 0j)
            QcB = 4.0 * pi * sqrt(rhoB / pi + 0j)
            Qcm = 0.022 * m  # critical q for given m-value
            g = Qcm / QcA
            QAB = QcA / sqrt(QcA**2 - QcB**2)
            b = beta * QAB**4 + 2.0 * beta * QAB**2 - 1
            N = int(abs(beta * ((g**2 * QAB**2 + 1) ** 2 - 1) - b))
            if N > 100000:
                raise ValueError(
                    "the number of bi-layers calculated is greater than 100 000, please check beta_sm parameter"
                )
        elif self.beta_sm < 0:
            # super-mirror from Hayter and Mook iterative method was selected, fixed number of layers
            if abs(self.beta_sm) > 1.0 or abs(self.beta_sm) < 0.6:
                raise ValueError("negative beta_sm should be between -1 and -0.6")
            N = int(abs(self.Repetitions))
        else:
            N = int(self.Repetitions)
        if name == "sigma":
            sigma_gradient = self.sigma_gradient
            # parameters for layers with roughness gradient
            par = [li.sigma for li in self.Layers]
            for i in range(1, N):
                if self.sigma_gtype == SigmaGradientType.rel_lin:
                    # linear increase of roughness to (1+sigma_gradient) times bottom roughness)
                    par += [li.sigma * (1.0 + (sigma_gradient + li.sigma_gradient) * i / (N - 1)) for li in self.Layers]
                elif self.sigma_gtype == SigmaGradientType.rel_sqrt:
                    # add roughness using rms
                    par += [
                        li.sigma * sqrt(1.0 + ((sigma_gradient + li.sigma_gradient) * i / (N - 1)) ** 2)
                        for li in self.Layers
                    ]
                elif self.sigma_gtype == SigmaGradientType.abs_lin:
                    # linear increase of roughness to bottom roughness + sigma_gradient)
                    par += [li.sigma + (sigma_gradient + li.sigma_gradient) * i / (N - 1) for li in self.Layers]
                elif self.sigma_gtype == SigmaGradientType.abs_sqrt:
                    # add roughness using rms
                    par += [
                        sqrt(li.sigma**2 + ((sigma_gradient + li.sigma_gradient) * i / (N - 1)) ** 2)
                        for li in self.Layers
                    ]
                else:
                    raise NotImplementedError("sigma_gtype must be in SigmaGradientType Enum")
        elif name == "d":
            d_gradient = self.d_gradient
            if self.beta_sm > 0:
                # layer thickness sequence based on calculated bi-layer thicknesses for super-mirror
                L_thicknesses = [li.d + 0.0 for li in self.Layers]
                D_start = sum(L_thicknesses)
                rel_thickness = [di / D_start for di in L_thicknesses]
                sm_scale = self.sm_scale
                idx = arange(N)
                D_SL = (
                    abs((2.0 * QAB * pi / QcA) / sqrt(sqrt(1 + (N - idx + b) / beta) - 1))
                    * (1.0 + d_gradient * idx / (N - 1))
                    * sm_scale
                )
                par = []
                for i in range(N):
                    par += [di * D_SL[i] for di in rel_thickness]
            elif self.beta_sm < 0:
                zeta = -self.beta_sm
                rho_all = [li.dens * (li.b.real + li.magn) * 1e-5 for li in self.Layers]
                rho1 = max(rho_all)  # in paper, SLD is called alpha
                rho2 = min(rho_all)
                # if two layers are chosen, use scale from olgorithm, else only change main 2 layers size
                L_thicknesses = [li.d + 0.0 for li in self.Layers]
                D_start = sum(L_thicknesses)
                idx1, idx2 = rho_all.index(rho1), rho_all.index(rho2)
                rest_size = D_start - L_thicknesses[idx1] - L_thicknesses[idx2]
                # layer thicknesses according to Hayter+Mook algorithm
                D_SL = sm_layers(rho1, rho2, N, zeta)
                # sign on repetitions defines direction of multilayer
                if self.Repetitions > 0:
                    D_SL.reverse()
                par = []
                sm_scale = self.sm_scale
                for i, (d1, d2) in enumerate(D_SL):
                    # variation from user parameters
                    var_scale = (1.0 + d_gradient * i / (N - 1)) * sm_scale
                    # scale the size of main two layer but remove residual layer thicknesses
                    total_bilayer = d1 + d2
                    scaled_bilayer = (total_bilayer - rest_size) / total_bilayer
                    for i, di in enumerate(L_thicknesses):
                        if i == idx1:
                            par.append(d1 * scaled_bilayer * var_scale)
                        elif i == idx2:
                            par.append(d2 * scaled_bilayer * var_scale)
                        else:
                            par.append(di * var_scale)
            else:
                # parameters for layers with thickness gradient
                par = []
                for i in range(N):
                    par += [
                        li.d * (1.0 - (d_gradient + li.d_gradient) * (1.0 / 2.0 - float(i) / N)) for li in self.Layers
                    ]
        elif name == "dens":
            dens_gradient = self.dens_gradient
            # parameters for layers with roughness gradient
            par = []
            for i in range(N):
                par += [
                    li.dens * (1.0 - dens_gradient * (1.0 / 2.0 - float(i) / self.Repetitions)) for li in self.Layers
                ]
        else:
            par = [getattr(li, name) for li in self.Layers] * N
        return par


class InhomType(AltStrEnum):
    gauss = "gauss"
    semi_gauss = "semi-gauss"
    empiric = "empiric PLD"


@dataclass
class Sample(refl.SampleBase):
    """
    Describe global sample by listing ambient, substrate and layer parameters.

    ``Stacks``
       A ``list`` consiting of ``Stack``s in the stacks the first item is
       the layer closest to the bottom
    ``Ambient``
       A ``Layer`` describing the Ambient (enviroment above the sample).
       Only the scattering lengths and density of the layer is used.
    ``Substrate``
       A ``Layer`` describing the substrate (enviroment below the sample).
       Only the scattering lengths, density and roughness of the layer is
       used.
    ``sigma_inhom``
       Width of the thickness distribution
    ``lscale_inhom``
       For empirical PLD model this defines the width of the increas peak
    ``flatwidth_inhom``
       For empirical PLD model this defines the trnsition from flat to increased probability.
    ``steps_inhom``
       Number of simulations to be performed to sample the thickness distribution model
    ``type_inhom``
       Function for the thickness probability. Either symmetric gaussian, semi-gaussian which does not have any probability
       for a higher thickness and an empirical model developed from PLD samples.
    ``crop_sld``
       Useful for multilayers with a very large number of repetitions. Only keeps crop_sld number of layers
       on top and bottom and removes all in the center for the SLD plot. The removed layers are replace with
       one layer of empty space. If the parameter is negative this gap can be replaced by a "peak" that
       separates top and bottom, this does invalidate the x-axis of the SLD plot, though.

    The empricial PLD model is base of simulations from x-ray line focus plume shapes to get thickness
    distribution probabilities and reduced to a smaller number of parameters.
    See PhD thesis *Multiferroicity in oxide thin films and heterostructures, A.Glavic, RWTH Aachen, (2012)* for a
    more detailed description of the empirical PLD model and an example application.
    """

    Stacks: List[Stack] = field(default_factory=list)
    Ambient: Layer = field(default_factory=Layer)
    Substrate: Layer = field(default_factory=Layer)
    crop_sld: int = 200
    sigma_inhom: float = 0.0
    lscale_inhom: float = 0.9
    flatwidth_inhom: float = 0.3
    steps_inhom: int = 20
    type_inhom: InhomType = "empiric PLD"

    _layer_parameter_class = LayerParameters


def Specular(TwoThetaQz, sample: Sample, instrument: Instrument):
    """
    The model function. Averadging the intensities for different
    layer thicknesses as found for e.g. large PLD samples.
    """
    global __xlabel__

    Q, TwoThetaQz, weight = spec_nx.resolution_init(TwoThetaQz, instrument)
    # often an issue with resolution etc. so just replace Q values < q_limit
    # if any(Q < q_limit):
    #    raise ValueError('The q vector has to be above %.1e, please verify all your x-axis points fulfill this criterion, including possible resolution smearing.'%q_limit)
    Q = np.maximum(Q, q_limit)

    # for better performance, perform all simulations without resolution convolution or footprint correction
    restype = instrument.restype
    foottype = instrument.footype
    Ibkg = instrument.Ibkg
    instrument.restype = ResType("no conv")
    instrument.footype = FootType("no corr")
    instrument.Ibkg = 0.0
    sampcall = deepcopy(sample)

    # average thicknesses before inhomogeniety average
    d0 = [array([Layer.d for Layer in Stack.Layers]) for Stack in sample.Stacks]
    sigma_d = sample.sigma_inhom * 0.01  # Inhomogeniety in \% (gamma for type 2)
    lorentz_scale = sample.lscale_inhom
    flat_width = np.maximum(1e-4, sample.flatwidth_inhom * 0.01)
    type_inhom = sample.type_inhom
    # Define the thicknesses to calculate and their propability
    if sigma_d == 0:  # no inhomogeniety
        d_fact = array([1.0])
        P = array([1.0])
    elif type_inhom == InhomType.semi_gauss:  # half gaussian shape inhomogeniety
        d_fact = 1.0 + linspace(-2.0 * sigma_d, 0, sample.steps_inhom)
        P = exp(-0.5 * (d_fact - sigma_d - 1.0) ** 2 / sigma_d**2)
        P /= P.sum()
        mean_d = (P * d_fact).sum()
        d_fact += 1.0 - mean_d
    elif type_inhom == InhomType.empiric:  # inhomogeniety of a PLD line focus, better approximation
        d_fact = 1.0 + linspace(-1.0 * max(2.0 * sigma_d, flat_width), 0, sample.steps_inhom)
        Pg = where(d_fact > flat_width, lorentz_scale * 1.0 / (1.0 + ((d_fact - 1.0) / sigma_d) ** 2), 0.0)
        Pf = (1.0 - lorentz_scale) * where(d_fact > flat_width, 1.0, 0.0)
        P = Pg + Pf
        P /= P.sum()
        mean_d = (P * d_fact).sum()
        d_fact += 1.0 - mean_d
    else:  # gaussian inhomegeniety
        d_fact = 1.0 + linspace(-sigma_d, sigma_d, sample.steps_inhom)
        P = exp(-0.5 * (d_fact - 1.0) ** 2 / sigma_d**2)
        P /= P.sum()
    # list for reflectivities to average
    Rlist = []
    # Iterate over thicknesses
    for d_facti, Pi in zip(d_fact, P):
        di = [d_facti * d0i for d0i in d0]
        for i, Stack in enumerate(sampcall.Stacks):
            for j, Layer in enumerate(Stack.Layers):
                # Layer.setD(di[i][j])
                Layer.d = di[i][j]
        Rlist.append(Pi * spec_nx.Specular(TwoThetaQz, sampcall, instrument))
    R = array(Rlist).sum(axis=0)

    # apply resolution convolution and footprint correction to result
    instrument.restype = restype
    instrument.footype = foottype
    instrument.Ibkg = Ibkg

    # footprint correction
    foocor = spec_nx.footprintcorr(Q, instrument)
    # resolution correction
    R = spec_nx.resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

    __xlabel__ = spec_nx.__xlabel__
    return R + instrument.Ibkg



def PolSpecular(TwoThetaQz, p1, p2, F1, F2, sample, instrument):
    """
    Specular reflectivity of polarized measurement with finite polarization.
    The polarization parameters are in accordance to the definition used in
    A.R. Wildes publication, Review of Scientific Instruments 70, 11 (1999)
    https://doi.org/10.1063/1.1150060

    pol_params = (p1, p2, f1, f2)
    p1: polarizer efficiency, 0: 100% spin-up, 1: 100% spin-down, 0.5: unpolarized
    p2: analyzer efficiency,  0: 100% spin-up, 1: 100% spin-down, 0.5: unpolairzed
    F1/F2: Flipper efficienty, 0: 100% efficient, 1: no flipping

    # BEGIN Parameters
    TwoThetaQz data.x
    p1 0.
    p2 0.
    F1 0.
    F2 0.
    # END Parameters

    """
    inst_pol = instrument.pol
    if not inst_pol in spec_nx.POL_CHANNELS:
        raise ValueError(f"Instrument polarization as to be one of {spec_nx.POL_CHANNELS}.")
    if instrument.probe != Probe.npolsf:
        raise ValueError("Polarization corrected simulation requires probe to be 'neutron pol spin flip'")

    instrument.pol = "uu"
    uu = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = "dd"
    dd = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = "ud"
    ud = Specular(TwoThetaQz, sample, instrument)
    du = ud
    instrument.pol = inst_pol

    P = get_pol_matrix(p1, p2, F1, F2)
    Pline = P[spec_nx.POL_CHANNELS.index(instrument.pol)]
    I = Pline[:, newaxis] * np.vstack([uu, ud, du, dd])
    return I.sum(axis=0)


SLD_calculations = spec_nx.SLD_calculations

SimulationFunctions = {"Specular": Specular, "PolSpecular": PolSpecular, "SLD": SLD_calculations}

Sample.setSimulationFunctions(SimulationFunctions)


class TestSpecInhom(ModelTestCase):
    # TODO: currently this only checks for raise conditions in the code above, check of results should be added

    def test_spec_xray(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)])],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=1.54,
        )
        with self.subTest("x-ray tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)

        # resolution corrections
        with self.subTest("x-ray q-res-fast"):
            instrument.restype = ResType.fast_conv
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-rel"):
            instrument.restype = ResType.fast_conv_rel
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-var"):
            instrument.restype = ResType.fast_conv_var
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full"):
            instrument.restype = ResType.full_conv_abs
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full-rel"):
            instrument.restype = ResType.full_conv_rel
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-gauss"):
            instrument.restype = resolution_module.GaussianResolution(sigma=0.001)
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-trapezoid"):
            instrument.restype = resolution_module.TrapezoidResolution(inner_width=0.001, outer_width=0.002)
            instrument.restype.get_weight_example()
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-wrong"):
            instrument.restype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)
        instrument.restype = ResType.none

        # footprint corrections
        with self.subTest("x-ray q-footprint-square"):
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        instrument.coords = Coords.tth
        with self.subTest("x-ray tth-footprint-square"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss"):
            instrument.footype = FootType.gauss
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss-offset"):
            instrument.footype = GaussianBeamOffset(sigma=0.5, offset=0.1)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-square-offset"):
            instrument.footype = SquareBeamOffset(width=0.1, offset=0.05)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-trapezoid"):
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.2)
            Specular(self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.1)
            Specular(self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.0, outer_width=0.2)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-wrong"):
            instrument.footype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)

    def test_spec_neutron(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0)])],
            Ambient=Layer(b=1e-7, dens=0.1),
            Substrate=Layer(b=4e-6, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.neutron,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        with self.subTest("neutron tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("neutron q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof q-footprint"):
            instrument.probe = Probe.ntof
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof q-footprint"):
            instrument.footype = FootType.none
            instrument.restype = ResType.full_conv_rel
            Specular(self.qz, sample, instrument)
        instrument.incangle = 0.5
        instrument.restype = ResType.none
        with self.subTest("neutron-pol++ q"):
            instrument.probe = Probe.npol
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-pol-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-polsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpol++ q"):
            instrument.probe = Probe.ntofpol
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpol-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpolsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)

        with self.subTest("neutron tthoffset"):
            instrument.probe = Probe.neutron
            instrument.tthoff = 0.1
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof tthoffset"):
            instrument.probe = Probe.ntof
            instrument.tthoff = 0.1
            Specular(self.qz, sample, instrument)

        with self.subTest("neutron-sf++ q"):
            instrument.probe = Probe.npolsf
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf+- q"):
            instrument.pol = Polarization.up_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf-+ q"):
            instrument.pol = Polarization.down_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sfsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-polspec q"):
            instrument.pol = Polarization.up_up
            PolSpecular(self.qz, 0.01, 0.01, 0.01, 0.01, sample, instrument)

    def test_sld(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=2e-5 + 1e-7j, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0)])],
            Ambient=Layer(b=1e-7, dens=0.1),
            Substrate=Layer(b=4e-6, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        with self.subTest("sld xray"):
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron"):
            instrument.probe = Probe.neutron
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron pol"):
            instrument.probe = Probe.npolsf
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron pol2"):
            sample.Stacks[0].Layers[0].magn_ang = 0.0
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron crop"):
            instrument.probe = Probe.neutron
            sample.crop_sld = 15
            sample.Stacks[0].Repetitions = 100
            SLD_calculations(None, None, sample, instrument)
            sample.crop_sld = -15
            sample.Stacks[0].Repetitions = 100
            SLD_calculations(None, None, sample, instrument)

    def test_inhom_types(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)])],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
            sigma_inhom=0.5,
            type_inhom=InhomType.gauss,
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.q,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=1.54,
        )
        with self.subTest("inhom gauss"):
            Specular(self.qz, sample, instrument)
        with self.subTest("inhom semi-gauss"):
            sample.type_inhom = InhomType.semi_gauss
            Specular(self.qz, sample, instrument)
        with self.subTest("inhom empiric"):
            sample.type_inhom = InhomType.empiric
            Specular(self.qz, sample, instrument)

    def test_gradients(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)], Repetitions=10)],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
            sigma_inhom=0.5,
            type_inhom=InhomType.gauss,
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.q,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=1.54,
        )
        with self.subTest("stack d gradient"):
            sample.Stacks[0].d_gradient = 1.0
            Specular(self.qz, sample, instrument)
        with self.subTest("stack sigma gradient abs_lin"):
            sample.Stacks[0].d_gradient = 0.0
            sample.Stacks[0].sigma_gradient = 5.0
            sample.Stacks[0].sigma_gtype = SigmaGradientType.abs_lin
            Specular(self.qz, sample, instrument)
        with self.subTest("stack sigma gradient abs_sqrt"):
            sample.Stacks[0].sigma_gradient = 5.0
            sample.Stacks[0].sigma_gtype = SigmaGradientType.abs_sqrt
            Specular(self.qz, sample, instrument)
        with self.subTest("stack sigma gradient rel_lin"):
            sample.Stacks[0].sigma_gradient = 0.5
            sample.Stacks[0].sigma_gtype = SigmaGradientType.rel_lin
            Specular(self.qz, sample, instrument)
        with self.subTest("stack sigma gradient rel_sqrt"):
            sample.Stacks[0].sigma_gradient = 0.5
            sample.Stacks[0].sigma_gtype = SigmaGradientType.rel_sqrt
            Specular(self.qz, sample, instrument)

    def test_supermirrors(self):
        from .utils import bc
        sample = Sample(
            Stacks=[
                Stack(
                    Layers=[
                        Layer(d=50, sigma=2.0, b=bc.Ni, dens=8.908*0.602214/58.6934),
                        Layer(d=50, sigma=2.0, b=bc.Ti, dens=4.506*0.602214/47.867),
                    ],
                    Repetitions=2.0,
                )
            ],
            Ambient=Layer(),
            Substrate=Layer(b=bc.Si, dens=0.04995982812477898),
        )
        instrument = Instrument(
            probe=Probe.neutron,
            coords=Coords.q,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=1.54,
        )
        with self.subTest("analytic supermirror"):
            sample.Stacks[0].beta_sm = 0.75
            sample.Stacks[0].Repetitions = 2.5 # m-value
            Specular(self.qz, sample, instrument)
        with self.subTest("analytic supermirror"):
            sample.Stacks[0].beta_sm = -0.99
            sample.Stacks[0].Repetitions = 2000  # number layers
            Specular(self.qz, sample, instrument)

def standard_xray():
    """
        return the defied standard x-ray reflectivity to compare against other models
    """
    qz = linspace(0.01, 0.3, 15)
    return Specular(
        qz,
        Sample(
            Ambient=Layer(),
            Substrate=Layer(d=150.0, f=1e-5 + 1e-8j, dens=0.1),
            Stacks=[Stack(Layers=[Layer(d=150.0, f=2e-5 + 2e-8j, dens=0.1)])],
        ),
        Instrument(probe=Probe.xray, coords=Coords.q, wavelength=1.54, footype=FootType.none, restype=ResType.none),
    )



# run performance tests on spec_nx model with CPU numba to see performance of stack
# used to profile the model performance
import sys
import genx.models
sys.modules['models']=sys.modules['genx.models']
from genx.models.lib import paratt_numba, neutron_numba, paratt_cuda, neutron_cuda

import genx.models.spec_nx as model
from genx.models.utils import UserVars, fp, fw, bc, bw
from numpy import sqrt, linspace, pi

# BEGIN Instrument DO NOT CHANGE
inst = model.Instrument(footype = 'gauss beam',probe = 'neutron pol spin flip',beamw = 0.2,resintrange = 3,tthoff = 0.0,
                        pol = 'uu',wavelength = 4.4,respoints = 15,Ibkg = 0.0,I0 = 2,samplelen = 50.0,
                        restype = 'full conv and varying res.',coords = 'q',res = 0.001,incangle = 0.0)
fp.set_wavelength(inst.wavelength)
#Compability issues for pre-fw created gx files
try:
	 fw
except:
	pass
else:
	fw.set_wavelength(inst.wavelength)
# END Instrument

# BEGIN Sample DO NOT CHANGE
Amb = model.Layer(b = 0, d = 0.0, f = (1e-20+1e-20j), dens = 1.0, magn_ang = 0.0, sigma = 0.0, xs_ai = 0.0, magn = 0.0)
SiO = model.Layer(b = bc.Si + bc.O*2, d = 1205, f = (1e-20+1e-20j), dens = 0.026, magn_ang = 0.0, sigma = 2, xs_ai = 0.0, magn = 0.0)
Sub = model.Layer(b = bc.Si, d = 0.0, f = (1e-20+1e-20j), dens = 8/5.443**3, magn_ang = 0.0, sigma = 2, xs_ai = 0.0, magn = 0.0)

surf = model.Stack(Layers=[SiO], Repetitions = 100)

sample = model.Sample(Stacks = [surf], Ambient = Amb, Substrate = Sub)
# END Sample

# BEGIN Parameters DO NOT CHANGE
cp = UserVars()
cp.new_var('dtheta', 0.04)
cp.new_var('dlol', 0.007)
# END Parameters

def Sim(data):
    I = []
    # BEGIN Dataset 0 DO NOT CHANGE
    inst.setRes(sqrt((cp.dlol*data[0].x)**2 + (4*3.1415/4.4*cp.dtheta*pi/360)**2))
    I.append(sample.SimSpecular(data[0].x, inst))
    # END Dataset 0
    return I

class empty():
    pass
e=empty()
e.x=linspace(0.001, 0.3, 500)

print("Numba")
model.Paratt.ReflQ=paratt_numba.ReflQ
model.MatrixNeutron.Refl=neutron_numba.Refl
for i in range(100):
    model.Buffer.parameters=None
    Sim([e])
print("Cuda")
model.Paratt.ReflQ=paratt_cuda.ReflQ
model.MatrixNeutron.Refl=neutron_cuda.Refl
for i in range(100):
    model.Buffer.parameters=None
    Sim([e])



"""
Constants used in various computations.
"""

r_e = 2.8179403227e-5  # classical electron radius [Å] ( 1/(4 pi epsilon0) * (e_charge**2)/(m_e*c**2) )
N_a = 6.022_140_76e23  # 1/mol

AA_to_eV = 12_398.41984 # x-ray energy conversion [eV/Å] ( h * c / 1Å)

M_to_SLD = 2.853e-6  # m/A/Å² conversion M [A/m] to SLD [Å^-2]
ApM_to_emucc = 1e-3  # conversion M [A/m] to M [emu/cm³]
Memu_to_SLD = M_to_SLD * ApM_to_emucc
muB = 9.2740101e-24  # Bohr magneton [A·m^2]

muB_to_SL = muB * M_to_SLD * 1e24  # magnetic scattering length for a moment of 1 Bohr magneton
T_to_SL = 2.31605e-6  # 1/T/Á² magnetic field in Tesla to scattering length
AAm2_to_emucc = 1e-6 / Memu_to_SLD  # conversion of SLD [Å/m²] to M [emu/cm³]

MASS_DENSITY_CONVERSION = 1e-24 * N_a  # g/cm³-> u/Å³ - used in converting mass density for SLDs



import math

from functools import reduce

from numpy import array, complex128, cos, exp, newaxis, pi, sqrt

from genx.core.custom_logging import iprint


# "Ordinary" implementaion of Parrats recursion formula
# theta-vector, lamda- can be a vector,n-1Dvector, d-1Dvector, sigma-1Dvector
def Refl(theta, lamda, n, d, sigma, return_int=True):
    d = d[1:-1]
    sigma = sigma[:-1]
    # Length of k-vector in vaccum
    k = 2 * math.pi / lamda
    # Calculates the wavevector in each layer
    Qj = 2 * n[-1] * k * sqrt(n[:, newaxis] ** 2 / n[-1] ** 2 - cos(theta * math.pi / 180) ** 2)
    # Fresnel reflectivity for the interfaces
    rp = (Qj[1:] - Qj[:-1]) / (Qj[1:] + Qj[:-1]) * exp(-Qj[1:] * Qj[:-1] / 2 * sigma[:, newaxis] ** 2)
    # Ignoring the top and bottom layer for the calc.
    p = exp(1.0j * d[:, newaxis] * Qj[1:-1])
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp = array(list(map(lambda x, y: [x, y], rp[1:], p)))

    # Paratt's recursion formula
    def formula(rtot, rint):
        return (rint[0] + rtot * rint[1]) / (1 + rtot * rint[0] * rint[1])

    # Implement the recursion formula
    r = reduce(formula, rpp, rp[0])

    if return_int:
        return abs(r) ** 2
    else:
        return r


# "Ordinary" implementaion of Parrats recursion formula
# Q-vector,n-1Dvector, d-1Dvector, sigma-1Dvector
def ReflQ(Q, lamda, n, d, sigma, return_int=True):
    # Length of k-vector in vaccum
    d = d[1:-1]
    sigma = sigma[:-1]
    Q0 = 4 * pi / lamda
    Q = Q.astype(complex128)
    # Calculates the wavevector in each layer
    Qj = sqrt((n[:, newaxis] ** 2 - n[-1] ** 2) * Q0**2 + n[-1] ** 2 * Q**2)
    # Fresnel reflectivity for the interfaces
    rp = (Qj[1:] - Qj[:-1]) / (Qj[1:] + Qj[:-1]) * exp(-Qj[1:] * Qj[:-1] / 2 * sigma[:, newaxis] ** 2)
    # print rp.shape #For debugging
    # print d.shape
    # print Qj[1:-1].shape
    p = exp(1.0j * d[:, newaxis] * Qj[1:-1])  # Ignoring the top and bottom layer for the calc.
    # print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp = array(list(map(lambda x, y: [x, y], rp[1:], p)))

    # print rpp.shape
    # Paratt's recursion formula
    def formula(rtot, rint):
        return (rint[0] + rtot * rint[1]) / (1 + rtot * rint[0] * rint[1])

    # Implement the recursion formula
    r = reduce(formula, rpp, rp[0])
    # print r.shape
    # return the reflectivity
    if return_int:
        return abs(r) ** 2
    else:
        return r


# paratts algorithm for n as function of lamda or theta
def Refl_nvary2(theta, lamda, n_vector, d, sigma, return_int=True):
    d = d[1:-1]
    sigma = sigma[:-1]
    # Length of k-vector in vaccum
    k = 2 * math.pi / lamda
    # Calculates the wavevector in each layer
    # print n_func
    # ss=transpose((sin(theta[:,newaxis]*pi/180.0)/lamda)*ones(len(n_func)))
    # print ss.shape
    # print theta.shape
    # print len(n_func)
    # n=array(map(lambda f,val:f(val),n_func,ss))
    n = n_vector
    # print n
    Qj = 2 * n[-1] * k * sqrt(n**2 / n[-1] ** 2 - cos(theta * math.pi / 180) ** 2)
    # print sigma.shape, Qj.shape
    # Fresnel reflectivity for the interfaces
    rp = (Qj[1:] - Qj[:-1]) / (Qj[1:] + Qj[:-1]) * exp(-Qj[1:] * Qj[:-1] / 2 * sigma[:, newaxis] ** 2)
    # print rp.shape #For debugging
    # print d.shape
    # print Qj[1:-1].shape
    p = exp(1.0j * d[:, newaxis] * Qj[1:-1])  # Ignoring the top and bottom layer for the calc.
    # print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp = array(list(map(lambda x, y: [x, y], rp[1:], p)))

    # print rpp.shape
    # Paratt's recursion formula
    def formula(rtot, rint):
        return (rint[0] + rtot * rint[1]) / (1 + rtot * rint[0] * rint[1])

    # Implement the recursion formula
    r = reduce(formula, rpp, rp[0])
    if return_int:
        return abs(r) ** 2
    else:
        return r


def reflq_kin(q, lamda, n, d, sigma, correct_q=True, return_int=True):
    """Calculates the reflectivity in the kinematical approximation"""
    d = d[:-1]
    d[0] = 0
    z = d.sum() - d.cumsum()
    sigma = sigma[:-1]
    q0 = 4 * pi / lamda
    # Kinematical reflectivity for the interfaces
    iprint(n.shape, len(n.shape))
    if len(n.shape) == 1:
        if correct_q:
            # The internal wave vector calacuted with the thickness averaged refractive index.
            n_mean = (n[1:-1] * d[1:] / d.sum()).sum()
            q_corr = sqrt((n_mean**2 - n[-1] ** 2) * q0**2 + (n[-1] * q) ** 2)
        else:
            q_corr = q
        rp = (n[:-1] - n[1:])[:, newaxis] * exp(-((q_corr * sigma[:, newaxis]) ** 2) / 2)
    else:
        if correct_q:
            # The internal wave vector calacuted with the thickness averaged refractive index.
            n_mean = (n[1:-1] * d[1:][:, newaxis] / d.sum()).sum(axis=0)
            q_corr = sqrt((n_mean**2 - n[-1] ** 2) * q0**2 + (n[-1] * q) ** 2)
        else:
            q_corr = q
        rp = (n[:-1] - n[1:]) * exp(-((q_corr * sigma[:, newaxis]) ** 2) / 2)
    p = exp(1.0j * z[:, newaxis] * q_corr)

    r = (rp * p).sum(axis=0) * q0**2 / q_corr**2 / 2.0

    if return_int:
        return abs(r) ** 2
    else:
        return r


def reflq_pseudo_kin(q, lamda, n, d, sigma, return_int=True):
    """Calculates the reflectivity in a pseudo kinematical approximation.
    The mean refractive index of the film is simulated with the single reflection approximation and the deviation from
    the mean is simulated with the kinematical approximation.
    """
    d = d[:-1]
    d[0] = 0
    z = d.sum() - d.cumsum()
    sigma = sigma[:-1]
    q0 = 4 * pi / lamda
    # Q = Q.astype(complex128)
    # The internal wave vector calacuted with the thickness averaged refractive index.
    n_mean = (n[1:-1] * d[1:] / d.sum()).sum()
    q_corr = sqrt((n_mean**2 - n[-1] ** 2) * q0**2 + n[-1] ** 2 * q**2)
    q_sub = sqrt((n[0] ** 2 - n[-1] ** 2) * q0**2 + n[-1] ** 2 * q**2)
    q_amb = n[-1] * q
    # Top interface
    rp_top = (q_corr - q_amb) / (q_corr + q_amb) * exp(-q_corr * q / 2 * sigma[-1] ** 2)
    rp_sub = (q_sub - q_corr) / (q_sub + q_corr) * exp(-q_sub * q_corr / 2 * sigma[0] ** 2)
    # rp_top = -(n[-1] - n_mean)*exp(-(q_corr*sigma[-1])**2/2)*q0**2/q_corr**2/2.
    # rp_sub = -(n_mean - n[0])*exp(-(q_corr*sigma[0])**2/2)*q0**2/q_corr**2/2.
    # Kinematical reflectivity for the interfaces
    n_diff = n - n_mean
    n_diff[0] = 0
    n_diff[-1] = 0
    rp = (n_diff[:-1] - n_diff[1:])[:, newaxis] * exp(-((q_corr * sigma[:, newaxis]) ** 2) / 2)
    p = exp(1.0j * z[:, newaxis] * q_corr)

    r_kin = (rp * p).sum(axis=0) * q0**2 / q_corr**2 / 2.0
    r_sra = rp_top + rp_sub * exp(1.0j * d.sum() * q_corr)
    r = r_kin + r_sra

    if return_int:
        return abs(r) ** 2
    else:
        return r


def reflq_sra(q, lamda, n, d, sigma, return_int=True):
    """Single reflection approximation calculation of the reflectivity"""
    # Length of k-vector in vaccum
    d = d[1:-1]
    sigma = sigma[:-1]
    q0 = 4 * pi / lamda
    # Calculates the wavevector in each layer
    if len(n.shape) == 1:
        qj = sqrt((n[:, newaxis] ** 2 - n[-1] ** 2) * q0**2 + (n[-1] * q) ** 2)
    else:
        qj = sqrt((n**2 - n[-1] ** 2) * q0**2 + (n[-1] * q) ** 2)
    # Fresnel reflectivity for the interfaces
    rp = (qj[:-1] - qj[1:]) / (qj[1:] + qj[:-1]) * exp(-qj[1:] * qj[:-1] / 2 * sigma[:, newaxis] ** 2)
    # The wave does not transverse the ambient and substrate - ignoring them
    # Also, the wave travels from top -> bottom, the array has the first element as the substrate
    # - need to reverse the order.
    phaseterm = (d[:, newaxis] * qj[1:-1])[::-1].cumsum(axis=0)[::-1]
    p = exp(1.0j * phaseterm)
    # Adding the first interface (top -> last in array) since p is not calculated for that layer (p = 1)
    r = rp[-1] + (rp[:-1] * p).sum(axis=0)
    if return_int:
        return abs(r) ** 2
    else:
        return r


from . import USE_NUMBA

if USE_NUMBA:
    # try to use numba to speed up the calculation intensive functions:
    try:
        from .paratt_numba import Refl, Refl_nvary2, ReflQ
    except Exception as e:
        iprint("Could not use numba, no speed up from JIT compiler:\n" + str(e))



import numpy as np

from genx.core.custom_logging import iprint

_ctype = np.complex128


def refl(kj, kjm1):
    return (kj + kjm1) / 2 / kjm1


def ass_X_test(k_b, k_u, k_l, dd_u, dd_l, sigma, sigma_l, sigma_u, dtype=_ctype):
    """Just a simpler version of ass_X to locate an error somewhere..."""
    k_jm1_b = k_b[..., 1:]
    k_j_b = k_b[..., :-1]
    k_jm1_l = k_l[..., 1:]
    k_j_u = k_u[..., :-1]
    sigma_jm1_l = sigma_l[..., 1:]
    sigma_j_b = sigma[..., :-1]
    sigma_j_u = sigma_u[..., :-1]
    dd_jm1_l = dd_l[..., 1:]
    dd_j_u = dd_u[..., :-1]

    iprint(dd_j_u, dd_jm1_l)

    # Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    # print 'r_j_b', r_j_b
    r_jm1_l = refl(k_jm1_l, k_jm1_b)
    # print 'r_jm1_l', r_jm1_l
    r_j_u = refl(k_j_u, k_jm1_l)
    # print 'r_j_u', r_j_u
    p_jm1_l = np.exp(-1.0j * dd_jm1_l * k_jm1_l)
    p_j_u = np.exp(-1.0j * dd_j_u * k_j_u)

    # Defining the X matrix
    X = np.empty((2, 2) + k_j_b.shape, dtype=dtype)

    X[0, 0] = p_jm1_l * r_j_u * p_j_u
    X[0, 1] = -(p_jm1_l * r_j_u - p_jm1_l) / p_j_u
    X[1, 0] = -((r_j_u - 1) * p_j_u) / p_jm1_l
    X[1, 1] = r_j_u / (p_jm1_l * p_j_u)

    return X


def ass_X_interfacelayer4(k_b, k_u, k_l, dd_u, dd_l, sigma, sigma_l, sigma_u, dtype=_ctype):
    """Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.4
    """
    # print dd_l, dd_u
    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[..., 1:]
    k_j_b = k_b[..., :-1]
    k_jm1_l = k_l[..., 1:]
    k_j_u = k_u[..., :-1]
    sigma_jm1_l = sigma_l[..., 1:]
    sigma_j_b = sigma[..., :-1]
    sigma_j_u = sigma_u[..., :-1]
    dd_jm1_l = dd_l[..., 1:]
    dd_j_u = dd_u[..., :-1]

    # Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    r_jm1_l = refl(k_jm1_l, k_jm1_b)
    r_j = refl(k_j_u, k_jm1_l)
    p_jm1_l = np.exp(-1.0j * dd_jm1_l * k_jm1_l)
    p_j_u = np.exp(-1.0j * dd_j_u * k_j_u)
    # p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    # Defining the X matrix
    X = np.empty((2, 2) + k_j_b.shape, dtype=dtype)

    tmp1 = k_jm1_b**2
    tmp2 = sigma_jm1_l**2
    tmp3 = -(tmp1 * tmp2) / 2
    tmp4 = -k_jm1_b * tmp2 * k_jm1_l
    tmp5 = -(tmp2 * k_jm1_l**2) / 2
    tmp6 = sigma_j_b**2
    tmp7 = -(tmp1 * tmp6) / 2
    tmp8 = k_j_b**2
    tmp9 = -(tmp6 * tmp8) / 2
    tmp10 = sigma_j_u**2
    tmp11 = -(tmp8 * tmp10) / 2
    tmp12 = -k_j_b * tmp10 * k_j_u
    tmp13 = -(tmp10 * k_j_u**2) / 2
    tmp14 = np.exp(tmp13 + tmp12 + tmp11 + tmp9 + tmp7 + tmp5 + tmp4 + tmp3)
    tmp15 = 1 / p_jm1_l
    tmp16 = 1 / p_j_u
    tmp17 = 2 * k_jm1_b * tmp2 * k_jm1_l
    tmp18 = k_jm1_b * tmp6 * k_j_b
    tmp21 = np.exp(tmp18)
    tmp19 = np.exp(tmp17) * tmp21
    tmp20 = p_jm1_l**2
    # tmp21=np.exp(tmp18)
    tmp22 = -tmp21
    tmp23 = 2 * k_j_b * tmp10 * k_j_u
    # tmp24=np.exp(tmp23+tmp18)
    # tmp24=np.exp(tmp23)*tmp21
    # tmp25=np.exp(tmp23+tmp18+tmp17)
    tmp26 = p_j_u**2
    # tmp27=np.exp(tmp13+tmp12+tmp11+tmp9-k_jm1_b*tmp6*k_j_b+tmp7+tmp5+tmp4+tmp3)
    tmp27 = tmp14 / tmp21
    tmp28 = np.exp(tmp17)
    # tmp29=np.exp(tmp23+tmp17)
    tmp30 = np.exp(tmp23)
    tmp29 = tmp28 * tmp30
    tmp24 = tmp30 * tmp21
    tmp25 = tmp30 * tmp19
    tmp31 = -tmp20
    tmp32 = tmp21 * tmp20
    tmp33 = -tmp21 * tmp20
    tmp34 = tmp24 * tmp20

    X[0, 0] = (
        tmp14
        * tmp15
        * tmp16
        * (
            (
                (((tmp25 * tmp20 + tmp24) * r_jm1_l - tmp24) * r_j - tmp24 * r_jm1_l + tmp24) * tmp26
                + ((tmp19 * tmp20 + tmp21) * r_jm1_l + tmp22) * r_j
                - tmp19 * tmp20 * r_jm1_l
            )
            * r_j_b
            + ((tmp22 - tmp19 * tmp20) * r_jm1_l + tmp21) * r_j
            + tmp19 * tmp20 * r_jm1_l
        )
    )

    X[0, 1] = (
        -tmp27
        * tmp15
        * tmp16
        * (
            (
                (((tmp28 * tmp20 + 1) * r_jm1_l - 1) * r_j - r_jm1_l + 1) * tmp26
                + ((tmp29 * tmp20 + tmp30) * r_jm1_l - tmp30) * r_j
                - tmp29 * tmp20 * r_jm1_l
            )
            * r_j_b
            + (((-tmp28 * tmp20 - 1) * r_jm1_l + 1) * r_j + r_jm1_l - 1) * tmp26
        )
    )

    X[1, 0] = (
        -tmp27
        * tmp15
        * tmp16
        * (
            (
                (((tmp30 * tmp20 + tmp29) * r_jm1_l - tmp30 * tmp20) * r_j - tmp29 * r_jm1_l) * tmp26
                + ((tmp20 + tmp28) * r_jm1_l + tmp31) * r_j
                - tmp20 * r_jm1_l
                + tmp20
            )
            * r_j_b
            + ((tmp31 - tmp28) * r_jm1_l + tmp20) * r_j
            + tmp20 * r_jm1_l
            + tmp31
        )
    )

    X[1, 1] = (
        tmp14
        * tmp15
        * tmp16
        * (
            (
                (((tmp32 + tmp19) * r_jm1_l + tmp33) * r_j - tmp19 * r_jm1_l) * tmp26
                + ((tmp34 + tmp25) * r_jm1_l - tmp24 * tmp20) * r_j
                - tmp24 * tmp20 * r_jm1_l
                + tmp34
            )
            * r_j_b
            + (((tmp33 - tmp19) * r_jm1_l + tmp32) * r_j + tmp19 * r_jm1_l) * tmp26
        )
    )

    return X


def ass_X_interfacelayer3(k_b, k_u, k_l, dd_u, dd_l, sigma, sigma_l, sigma_u, dtype=_ctype):
    """Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.3
    """

    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[..., 1:]
    k_j_b = k_b[..., :-1]
    k_jm1_l = k_l[..., 1:]
    k_j_u = k_u[..., :-1]
    sigma_jm1_l = sigma_l[..., 1:]
    sigma_j_b = sigma[..., :-1]
    sigma_j_u = sigma_u[..., :-1]
    dd_jm1_l = dd_l[..., 1:]
    dd_j_u = dd_u[..., :-1]

    iprint(dd_j_u, dd_jm1_l)

    # Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    iprint("r_j_b", r_j_b)
    r_jm1_l = refl(k_jm1_l, k_jm1_b)
    iprint("r_jm1_l", r_jm1_l)
    r_j_u = refl(k_j_u, k_jm1_l)
    iprint("r_j_u", r_j_u)
    p_jm1_l = np.exp(-1.0j * dd_jm1_l * k_jm1_l)
    p_j_u = np.exp(-1.0j * dd_j_u * k_j_u)
    # p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    # Defining the X matrix
    X = np.empty((2, 2) + k_j_b.shape, dtype=dtype)

    tmp1 = k_jm1_b**2
    tmp2 = sigma_jm1_l**2
    tmp3 = -(tmp1 * tmp2) / 2
    tmp4 = k_jm1_b * tmp2 * k_jm1_l
    tmp5 = -(tmp2 * k_jm1_l**2) / 2
    tmp6 = sigma_j_b**2
    tmp7 = -(tmp1 * tmp6) / 2
    tmp8 = k_jm1_b * tmp6 * k_j_b
    tmp9 = k_j_b**2
    tmp10 = -(tmp6 * tmp9) / 2
    tmp11 = sigma_j_u**2
    tmp12 = -(tmp9 * tmp11) / 2
    tmp13 = -k_j_b * tmp11 * k_j_u
    tmp14 = -(tmp11 * k_j_u**2) / 2
    tmp15 = np.exp(tmp14 + tmp13 + tmp12 + tmp10 + tmp8 + tmp7 + tmp5 + tmp4 + tmp3)
    tmp16 = 1 / p_j_u
    tmp17 = -k_jm1_b * tmp2 * k_jm1_l
    tmp18 = np.exp(tmp14 + tmp13 + tmp12 + tmp10 + tmp8 + tmp7 + tmp5 + tmp17 + tmp3)
    tmp19 = 1 / p_jm1_l
    tmp20 = k_j_b * tmp11 * k_j_u
    tmp21 = np.exp(tmp14 + tmp20 + tmp12 + tmp10 + tmp8 + tmp7 + tmp5 + tmp17 + tmp3)
    tmp22 = np.exp(tmp14 + tmp20 + tmp12 + tmp10 + tmp8 + tmp7 + tmp5 + tmp4 + tmp3)
    tmp23 = -k_jm1_b * tmp6 * k_j_b
    tmp24 = np.exp(tmp14 + tmp13 + tmp12 + tmp10 + tmp23 + tmp7 + tmp5 + tmp17 + tmp3)
    tmp25 = np.exp(tmp14 + tmp13 + tmp12 + tmp10 + tmp23 + tmp7 + tmp5 + tmp4 + tmp3)
    tmp26 = np.exp(tmp14 + tmp20 + tmp12 + tmp10 + tmp23 + tmp7 + tmp5 + tmp4 + tmp3)
    tmp27 = np.exp(tmp14 + tmp20 + tmp12 + tmp10 + tmp23 + tmp7 + tmp5 + tmp17 + tmp3)

    X[0, 0] = (
        tmp22 * p_jm1_l * r_jm1_l * r_j_b * p_j_u * r_j_u
        + tmp21 * tmp19 * r_jm1_l * r_j_b * p_j_u * r_j_u
        - tmp21 * tmp19 * r_j_b * p_j_u * r_j_u
        - tmp21 * tmp19 * r_jm1_l * p_j_u * r_j_u
        + tmp21 * tmp19 * p_j_u * r_j_u
        + tmp15 * p_jm1_l * r_jm1_l * r_j_b * tmp16 * r_j_u
        + tmp18 * tmp19 * r_jm1_l * r_j_b * tmp16 * r_j_u
        - tmp18 * tmp19 * r_j_b * tmp16 * r_j_u
        - tmp15 * p_jm1_l * r_jm1_l * tmp16 * r_j_u
        - tmp15 * p_jm1_l * r_jm1_l * r_j_b * tmp16
        - tmp18 * tmp19 * r_jm1_l * r_j_b * tmp16
        + tmp18 * tmp19 * r_j_b * tmp16
        + tmp15 * p_jm1_l * r_jm1_l * tmp16
    )

    X[0, 1] = (
        -tmp25 * p_jm1_l * r_jm1_l * r_j_b * p_j_u * r_j_u
        - tmp24 * tmp19 * r_jm1_l * r_j_b * p_j_u * r_j_u
        + tmp24 * tmp19 * r_j_b * p_j_u * r_j_u
        + tmp24 * tmp19 * r_jm1_l * p_j_u * r_j_u
        - tmp24 * tmp19 * p_j_u * r_j_u
        - tmp26 * p_jm1_l * r_jm1_l * r_j_b * tmp16 * r_j_u
        - tmp27 * tmp19 * r_jm1_l * r_j_b * tmp16 * r_j_u
        + tmp27 * tmp19 * r_j_b * tmp16 * r_j_u
        + tmp26 * p_jm1_l * r_jm1_l * tmp16 * r_j_u
        + tmp25 * p_jm1_l * r_jm1_l * r_j_b * p_j_u
        + tmp24 * tmp19 * r_jm1_l * r_j_b * p_j_u
        - tmp24 * tmp19 * r_j_b * p_j_u
        - tmp24 * tmp19 * r_jm1_l * p_j_u
        + tmp24 * tmp19 * p_j_u
    )

    X[1, 0] = (
        -tmp27 * p_jm1_l * r_jm1_l * r_j_b * p_j_u * r_j_u
        - tmp26 * tmp19 * r_jm1_l * r_j_b * p_j_u * r_j_u
        + tmp27 * p_jm1_l * r_j_b * p_j_u * r_j_u
        + tmp26 * tmp19 * r_jm1_l * p_j_u * r_j_u
        - tmp24 * p_jm1_l * r_jm1_l * r_j_b * tmp16 * r_j_u
        - tmp25 * tmp19 * r_jm1_l * r_j_b * tmp16 * r_j_u
        + tmp24 * p_jm1_l * r_j_b * tmp16 * r_j_u
        + tmp24 * p_jm1_l * r_jm1_l * tmp16 * r_j_u
        - tmp24 * p_jm1_l * tmp16 * r_j_u
        + tmp24 * p_jm1_l * r_jm1_l * r_j_b * tmp16
        + tmp25 * tmp19 * r_jm1_l * r_j_b * tmp16
        - tmp24 * p_jm1_l * r_j_b * tmp16
        - tmp24 * p_jm1_l * r_jm1_l * tmp16
        + tmp24 * p_jm1_l * tmp16
    )

    X[1, 1] = (
        tmp18 * p_jm1_l * r_jm1_l * r_j_b * p_j_u * r_j_u
        + tmp15 * tmp19 * r_jm1_l * r_j_b * p_j_u * r_j_u
        - tmp18 * p_jm1_l * r_j_b * p_j_u * r_j_u
        - tmp15 * tmp19 * r_jm1_l * p_j_u * r_j_u
        + tmp21 * p_jm1_l * r_jm1_l * r_j_b * tmp16 * r_j_u
        + tmp22 * tmp19 * r_jm1_l * r_j_b * tmp16 * r_j_u
        - tmp21 * p_jm1_l * r_j_b * tmp16 * r_j_u
        - tmp21 * p_jm1_l * r_jm1_l * tmp16 * r_j_u
        + tmp21 * p_jm1_l * tmp16 * r_j_u
        - tmp18 * p_jm1_l * r_jm1_l * r_j_b * p_j_u
        - tmp15 * tmp19 * r_jm1_l * r_j_b * p_j_u
        + tmp18 * p_jm1_l * r_j_b * p_j_u
        + tmp15 * tmp19 * r_jm1_l * p_j_u
    )

    return X


def ass_X_interfacelayer(k_b, k_u, k_l, dd_u, dd_l, sigma, sigma_l, sigma_u, dtype=_ctype):
    """Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.2
    """
    iprint(dd_l, dd_u)
    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[..., 1:]
    k_j_b = k_b[..., :-1]
    k_jm1_l = k_l[..., 1:]
    k_j_u = k_u[..., :-1]
    sigma_jm1_l = sigma_l[..., 1:]
    sigma_j_b = sigma[..., :-1]
    sigma_j_u = sigma_u[..., :-1]
    dd_jm1_l = dd_l[..., 1:]
    dd_j_u = dd_u[..., :-1]

    # Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    r_jm1_l = refl(k_jm1_l, k_jm1_b)
    r_j_u = refl(k_j_u, k_jm1_l)
    p_jm1_l = np.exp(-1.0j * dd_jm1_l * k_jm1_b)
    p_j_u = np.exp(-1.0j * dd_j_u * k_j_u)
    # p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    # Defining the X matrix
    X = np.empty((2, 2) + k_j_b.shape, dtype=dtype)

    tmp1 = k_jm1_b**2
    tmp2 = sigma_jm1_l**2
    tmp3 = -(tmp1 * tmp2) / 2
    tmp4 = -k_jm1_b * tmp2 * k_jm1_l
    tmp5 = -(tmp2 * k_jm1_l**2) / 2
    tmp6 = sigma_j_b**2
    tmp7 = -(tmp1 * tmp6) / 2
    tmp8 = k_j_b**2
    tmp9 = -(tmp6 * tmp8) / 2
    tmp10 = sigma_j_u**2
    tmp11 = -(tmp8 * tmp10) / 2
    tmp12 = -k_j_b * tmp10 * k_j_u
    tmp13 = -(tmp10 * k_j_u**2) / 2
    tmp14 = np.exp(tmp13 + tmp12 + tmp11 + tmp9 + tmp7 + tmp5 + tmp4 + tmp3)
    tmp15 = 1 / p_jm1_l
    tmp16 = 1 / p_j_u
    tmp17 = 2 * k_jm1_b * tmp2 * k_jm1_l
    tmp18 = k_jm1_b * tmp6 * k_j_b
    tmp21 = np.exp(tmp18)
    tmp19 = np.exp(tmp17) * tmp21
    tmp20 = p_jm1_l**2
    # tmp21=np.exp(tmp18)
    tmp22 = -tmp21
    tmp23 = 2 * k_j_b * tmp10 * k_j_u
    # tmp24=np.exp(tmp23+tmp18)
    # tmp24=np.exp(tmp23)*tmp21
    # tmp25=np.exp(tmp23+tmp18+tmp17)
    tmp26 = p_j_u**2
    # tmp27=np.exp(tmp13+tmp12+tmp11+tmp9-k_jm1_b*tmp6*k_j_b+tmp7+tmp5+tmp4+tmp3)
    tmp27 = tmp14 / tmp21
    tmp28 = np.exp(tmp17)
    # tmp29=np.exp(tmp23+tmp17)
    tmp30 = np.exp(tmp23)
    tmp29 = tmp28 * tmp30
    tmp24 = tmp30 * tmp21
    tmp25 = tmp30 * tmp19
    tmp31 = -tmp20
    tmp32 = tmp21 * tmp20
    tmp33 = -tmp21 * tmp20
    tmp34 = tmp24 * tmp20

    X[0, 0] = (
        tmp14
        * tmp15
        * tmp16
        * (
            (
                (((tmp25 * tmp20 + tmp24) * r_jm1_l - tmp24) * r_j_b - tmp24 * r_jm1_l + tmp24) * tmp26
                + ((tmp19 * tmp20 + tmp21) * r_jm1_l + tmp22) * r_j_b
                - tmp19 * tmp20 * r_jm1_l
            )
            * r_j_u
            + ((tmp22 - tmp19 * tmp20) * r_jm1_l + tmp21) * r_j_b
            + tmp19 * tmp20 * r_jm1_l
        )
    )

    X[0, 1] = (
        -tmp27
        * tmp15
        * tmp16
        * (
            (
                (((tmp28 * tmp20 + 1) * r_jm1_l - 1) * r_j_b - r_jm1_l + 1) * tmp26
                + ((tmp29 * tmp20 + tmp30) * r_jm1_l - tmp30) * r_j_b
                - tmp29 * tmp20 * r_jm1_l
            )
            * r_j_u
            + (((-tmp28 * tmp20 - 1) * r_jm1_l + 1) * r_j_b + r_jm1_l - 1) * tmp26
        )
    )

    X[1, 0] = (
        -tmp27
        * tmp15
        * tmp16
        * (
            (
                (((tmp30 * tmp20 + tmp29) * r_jm1_l - tmp30 * tmp20) * r_j_b - tmp29 * r_jm1_l) * tmp26
                + ((tmp20 + tmp28) * r_jm1_l + tmp31) * r_j_b
                - tmp20 * r_jm1_l
                + tmp20
            )
            * r_j_u
            + ((tmp31 - tmp28) * r_jm1_l + tmp20) * r_j_b
            + tmp20 * r_jm1_l
            + tmp31
        )
    )

    X[1, 1] = (
        tmp14
        * tmp15
        * tmp16
        * (
            (
                (((tmp32 + tmp19) * r_jm1_l + tmp33) * r_j_b - tmp19 * r_jm1_l) * tmp26
                + ((tmp34 + tmp25) * r_jm1_l - tmp24 * tmp20) * r_j_b
                - tmp24 * tmp20 * r_jm1_l
                + tmp34
            )
            * r_j_u
            + (((tmp33 - tmp19) * r_jm1_l + tmp32) * r_j_b + tmp19 * r_jm1_l) * tmp26
        )
    )

    return X


def ass_X_interfacelayer1(k_b, k_u, k_l, dd_u, dd_l, sigma, sigma_l, sigma_u, dtype=_ctype):
    """Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.1
    """

    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[..., 1:]
    k_j_b = k_b[..., :-1]
    k_jm1_l = k_l[..., 1:]
    k_j_u = k_u[..., :-1]
    sigma_jm1_l = sigma_l[..., 1:]
    sigma_j_b = sigma[..., :-1]
    sigma_j_u = sigma_u[..., :-1]
    dd_jm1_l = dd_u[..., 1:]
    dd_j_u = dd_u[..., :-1]

    # Defining the X matrix
    X = np.empty((2, 2) + k_j_b.shape, dtype=dtype)

    # Temporary variables to ease the calculation
    tmp1 = k_jm1_b * k_jm1_b
    tmp2 = sigma_jm1_l * sigma_jm1_l
    tmp3 = tmp1 * tmp2
    tmp4 = (2 * k_jm1_b * tmp2 + 2 * 1.0j * dd_jm1_l) * k_jm1_l
    tmp5 = k_jm1_l * k_jm1_l
    tmp6 = tmp2 * tmp5
    tmp7 = sigma_j_b * sigma_j_b
    tmp8 = tmp1 * tmp7
    tmp9 = k_j_b**2
    tmp10 = tmp7 * tmp9
    tmp11 = sigma_j_u * sigma_j_u
    tmp12 = tmp9 * tmp11
    tmp13 = (2 * k_j_b * tmp11 + 2 * 1.0j * dd_j_u) * k_j_u
    tmp14 = k_j_u * k_j_u
    tmp15 = tmp11 * tmp14
    tmp16 = 1 / np.exp((tmp15 + tmp13 + tmp12 + tmp10 + tmp8 + tmp6 + tmp4 + tmp3) / 2)
    tmp17 = 1 / k_jm1_b
    tmp18 = 1 / k_jm1_l
    tmp19 = 1 / k_j_u
    tmp20 = np.exp(2 * 1.0j * dd_jm1_l * k_jm1_l)
    tmp21 = -tmp20
    tmp22 = np.exp(2 * k_jm1_b * tmp2 * k_jm1_l)
    tmp23 = -tmp22
    tmp24 = tmp23 + tmp21
    tmp25 = k_jm1_b * tmp7 * k_j_b
    tmp26 = 2 * 1.0j * dd_j_u * k_j_u
    tmp27 = np.exp(tmp26 + tmp25)
    tmp28 = tmp22 + tmp20
    tmp29 = 2 * k_j_b * tmp11 * k_j_u
    tmp30 = np.exp(tmp29 + tmp25)
    tmp31 = tmp28 * tmp30
    tmp32 = tmp31 + tmp24 * tmp27
    tmp33 = tmp23 + tmp20
    tmp34 = tmp22 + tmp21
    tmp35 = tmp34 * tmp30
    tmp36 = tmp35 + tmp33 * tmp27
    tmp37 = tmp31 + tmp28 * tmp27
    tmp38 = tmp35 + tmp34 * tmp27
    tmp39 = 1 / np.exp((tmp15 + tmp13 + tmp12 + tmp10 + 2 * k_jm1_b * tmp7 * k_j_b + tmp8 + tmp6 + tmp4 + tmp3) / 2)
    tmp40 = np.exp(tmp13)
    tmp41 = tmp24 * tmp40
    tmp42 = tmp33 * tmp40
    tmp43 = tmp34 * tmp40
    tmp44 = tmp28 * tmp40
    tmp45 = np.exp(tmp4)
    tmp46 = tmp45 + 1
    tmp47 = np.exp(tmp26)
    tmp48 = tmp46 * tmp47
    tmp49 = -tmp45
    tmp50 = tmp49 - 1
    tmp51 = np.exp(tmp29)
    tmp52 = tmp50 * tmp51
    tmp53 = tmp45 - 1
    tmp54 = tmp53 * tmp47
    tmp55 = tmp49 + 1
    tmp56 = tmp55 * tmp51
    tmp57 = tmp50 * tmp47
    tmp58 = tmp55 * tmp47
    tmp59 = tmp53 * tmp51
    tmp60 = tmp46 * tmp51
    tmp61 = np.exp(tmp25)
    tmp62 = np.exp(tmp13 + tmp25)
    tmp63 = tmp46 * tmp62
    tmp64 = tmp63 + tmp50 * tmp61
    tmp65 = tmp53 * tmp62
    tmp66 = tmp65 + tmp55 * tmp61
    tmp67 = tmp63 + tmp46 * tmp61
    tmp68 = tmp65 + tmp53 * tmp61

    # Assembling the interface matrices
    X[0, 0] = (
        tmp16
        * tmp17
        * tmp18
        * tmp19
        * (
            (tmp32 * k_jm1_l + tmp36 * k_jm1_b) * tmp14
            + ((tmp37 * k_jm1_l + tmp38 * k_jm1_b) * k_j_b + tmp38 * tmp5 + tmp37 * k_jm1_b * k_jm1_l) * k_j_u
            + (tmp36 * tmp5 + tmp32 * k_jm1_b * k_jm1_l) * k_j_b
        )
    ) / 8

    X[0, 1] = (
        -(
            tmp39
            * tmp17
            * tmp18
            * tmp19
            * (
                ((tmp44 + tmp23 + tmp21) * k_jm1_l + (tmp43 + tmp23 + tmp20) * k_jm1_b) * tmp14
                + (
                    ((tmp44 + tmp22 + tmp20) * k_jm1_l + (tmp43 + tmp22 + tmp21) * k_jm1_b) * k_j_b
                    + (tmp42 + tmp23 + tmp20) * tmp5
                    + (tmp41 + tmp23 + tmp21) * k_jm1_b * k_jm1_l
                )
                * k_j_u
                + ((tmp42 + tmp22 + tmp21) * tmp5 + (tmp41 + tmp22 + tmp20) * k_jm1_b * k_jm1_l) * k_j_b
            )
        )
        / 8
    )

    X[1, 0] = (
        -(
            tmp39
            * tmp17
            * tmp18
            * tmp19
            * (
                ((tmp60 + tmp57) * k_jm1_l + (tmp59 + tmp58) * k_jm1_b) * tmp14
                + (
                    ((tmp60 + tmp48) * k_jm1_l + (tmp59 + tmp54) * k_jm1_b) * k_j_b
                    + (tmp56 + tmp58) * tmp5
                    + (tmp52 + tmp57) * k_jm1_b * k_jm1_l
                )
                * k_j_u
                + ((tmp56 + tmp54) * tmp5 + (tmp52 + tmp48) * k_jm1_b * k_jm1_l) * k_j_b
            )
        )
        / 8
    )

    X[1, 1] = (
        tmp16
        * tmp17
        * tmp18
        * tmp19
        * (
            (tmp64 * k_jm1_l + tmp66 * k_jm1_b) * tmp14
            + ((tmp67 * k_jm1_l + tmp68 * k_jm1_b) * k_j_b + tmp68 * tmp5 + tmp67 * k_jm1_b * k_jm1_l) * k_j_u
            + (tmp66 * tmp5 + tmp64 * k_jm1_b * k_jm1_l) * k_j_b
        )
    ) / 8

    return X



"""
===============================
:mod:`xrdml`  XRDML data loader
===============================

Loads the data from Philips XPert instrument.
"""

from xml.dom.minidom import parseString

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx

    from wx.lib.masked import NumCtrl
except ImportError:

    class void:
        pass

    wx = void()
    wx.Dialog = void


class Plugin(Template):
    wildcard = "*.xrdml"

    def __init__(self, parent):
        Template.__init__(self, parent)

    def CountDatasets(self, file_path):
        try:
            orso_datasets = ReadXpert(file_path)
        except Exception as e:
            return 1
        else:
            return len(orso_datasets)

    def LoadData(self, dataset, filename, data_id=0):
        """
        Loads the data from filename into the data_item_number.
        """
        try:
            datasets = ReadXpert(filename)
        except Exception as e:
            import traceback

            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n Error in ReadXpert:\n"
                + traceback.format_exc(),
            )
        else:
            ds = datasets[data_id]
            dataset.x_raw = ds[0]
            dataset.y_raw = ds[1]
            dataset.error_raw = ds[2]
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta["data_source"]["experiment"]["instrument"] = "XRDML"
            dataset.meta["data_source"]["experiment"]["probe"] = "x-ray"
            dataset.meta["data_source"]["measurement"]["scheme"] = "angle-dispersive"
            dataset.meta["data_source"]["sample"]["name"] = ds[3]
            dataset.meta["data_source"]["experiment"].update(ds[4])
            dataset.meta["data_source"]["measurement"].update(ds[5])


def ReadXpert(file_name):
    """
    Read the data of a philips X'Pert diffractometer file, exported as text files.
    """
    with open(file_name, "r") as fh:
        raw_data = fh.read()
    while len(raw_data) > 0 and raw_data[0] != "<":
        # some files are written using UTF-8 BOM format that has extra bytes before the text starts
        # this will ignore any characters at the beginning that are not "<", the tag opening for XML
        raw_data = raw_data[1:]
    xml_data = parseString(raw_data).firstChild

    # retrieve data
    try:
        sample_name = xml_data.getElementsByTagName("sample")[0].getElementsByTagName("name")[0].firstChild.nodeValue
    except AttributeError:
        sample_name = file_name.rsplit(".", 1)[0]
    datasets = []

    try:
        lamda = float(xml_data.getElementsByTagName("kAlpha1")[0].firstChild.nodeValue)
    except (IndexError, ValueError):
        lamda = 1.54  # Cu k-alpha

    for xml_scan in xml_data.getElementsByTagName("xrdMeasurement")[0].getElementsByTagName("scan"):
        meta_experiment = {}
        meta_measurement = {}
        scan = xml_scan.getElementsByTagName("dataPoints")[0]

        moving_positions = {}
        for motor in scan.getElementsByTagName("positions"):
            axis = motor.attributes["axis"].value
            if len(motor.getElementsByTagName("commonPosition")) == 0:
                start = float(motor.getElementsByTagName("startPosition")[0].firstChild.nodeValue)
                end = float(motor.getElementsByTagName("endPosition")[0].firstChild.nodeValue)
                moving_positions[axis] = (start, end)

        try:
            atten_factors = scan.getElementsByTagName("beamAttenuationFactors")[0].firstChild.nodeValue
            atten_factors = list(map(float, atten_factors.split()))
            atten = np.array(atten_factors)
        except IndexError:
            atten = 1.0
        try:
            meta_experiment["start_date"] = xml_scan.getElementsByTagName("startTimeStamp")[0].firstChild.nodeValue
        except IndexError:
            pass

        time = float(scan.getElementsByTagName("commonCountingTime")[0].firstChild.nodeValue)
        data_tags = scan.getElementsByTagName("intensities") + scan.getElementsByTagName("counts")
        data_tag = data_tags[0]
        data = list(map(float, data_tag.firstChild.nodeValue.split()))

        I = np.array(data)
        if data_tag.nodeName=="counts":
            dI = np.sqrt(I)
            I /= time / atten
            dI /= time / atten
        else:
            dI = np.sqrt(I*atten)
            I /= time
            dI /= time

        if "2Theta" in moving_positions:
            th = np.linspace(moving_positions["2Theta"][0], moving_positions["2Theta"][1], len(data)) / 2.0
        else:
            th = np.linspace(moving_positions["Omega"][0], moving_positions["Omega"][1], len(data))
        meta_measurement["instrument_settings"] = {
            "incident_angle": {"min": float(th.min()), "max": float(th.max()), "unit": "deg"},
            "wavelength": {"magnitude": lamda, "unit": "angstrom"},
        }
        datasets.append((2*th, I, dI, sample_name, meta_experiment, meta_measurement))
    return datasets



"""Library that implements an effective density model for composition 
profiles (as described by Tolan) and also a home constructed model for
magnetic moment profiles (element specific). 
"""

from scipy.special import erf
from numpy import r_, cumsum, where, sqrt, exp, pi, array, zeros, sin, arange, bitwise_not, bitwise_or, append, newaxis

def erf_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1 * z0 + sigma0 * (z0 + d)) / (sigma0 + sigma1)
    p = 0.5 * (1 + where(z <= eta, erf((z - z0) / sqrt(2.0) / sigma0), -erf((z - z0 - d) / sqrt(2.0) / sigma1)))
    return p


def erf_interf(z, sigma):
    return 0.5 + 0.5 * erf(z / sqrt(2.0) / sigma)


def _lin(z, sigma):
    p = zeros(z.shape)
    p = (z > sqrt(3) * sigma) * 1.0 + (abs(z) <= sqrt(3) * sigma) * (0.5 + z / 2 / sqrt(3) / sigma)
    return p


def lin_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1 * z0 + sigma0 * (z0 + d)) / (sigma0 + sigma1)
    p = where(z < eta, _lin(z - z0, sigma0), _lin(-(z - z0 - d), sigma1))
    return p


def _exp(z, sigma):
    p = (z <= 0) * (0.5 * exp(sqrt(2) * z / sigma)) + (z > 0) * (1 - 0.5 * exp(-sqrt(2) * z / sigma))
    return p


def exp_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1 * z0 + sigma0 * (z0 + d)) / (sigma0 + sigma1)
    p = where(z < eta, _exp(z - z0, sigma0), _exp(-(z - z0 - d), sigma1))
    return p


def _sin(z, sigma):
    a = pi / sqrt(pi**2 - 8) * sigma
    p = (z > a) * 1.0 + (abs(z) <= a) * (0.5 + 0.5 * sin(pi * z / 2 / a))
    return p


def sin_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1 * z0 + sigma0 * (z0 + d)) / (sigma0 + sigma1)
    p = where(z < eta, _sin(z - z0, sigma0), _sin(-(z - z0 - d), sigma1))
    return p


def compress_profile_old(z, p, delta_max):
    pnew = p.copy()
    znew = z.copy()
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        # print i
        i += 1
        index = abs(pnew[:-1] - pnew[1:]) > delta_max
        index[::2] = True
        index = r_[index, True]
        # print pnew.shape, index.shape
        pnew = pnew[index]
        znew = znew[index]

    return znew, pnew


def compress_profile(z, p, delta_max):
    """Compresses a profile by merging the neighbouring layers that
    has an density difference of maximum delta_max"""
    pnew = p.copy()
    znew = z.copy()
    inew = arange(len(p))
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        # print i
        i += 1
        index = array([True] * len(pnew))
        index[1:-1:2] = abs(pnew[:-2:2] - pnew[2::2]) > delta_max
        # print pnew.shape, index.shape
        pnew = pnew[index]
        znew = znew[index]
        inew = inew[index]
    # print inew
    inew = append(inew, inew[-1])
    pret = array([p[inew[i] : inew[i + 1] + 1].mean() for i in range(len(inew) - 1)])
    return znew, pret


def compress_profile2(z, p1, p2, delta_max):
    """Compresses a profile by merging the neighbouring layers that
    has an density difference of maximum delta_max"""
    p1new = p1.copy()
    p2new = p2.copy()
    znew = z.copy()
    inew = arange(len(p1))
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        # print i
        i += 1
        index = array([True] * len(p1new))
        index[1:-1:2] = bitwise_or(
            abs(p1new[:-2:2] - p1new[2::2]) > delta_max, abs(p2new[:-2:2] - p2new[2::2]) > delta_max
        )
        # print pnew.shape, index.shape
        p1new = p1new[index]
        p2new = p2new[index]
        znew = znew[index]
        inew = inew[index]
    # print inew
    inew = append(inew, inew[-1])
    p1ret = array([p1[inew[i] : inew[i + 1] + 1].mean() for i in range(len(inew) - 1)])
    p2ret = array([p2[inew[i] : inew[i + 1] + 1].mean() for i in range(len(inew) - 1)])
    return znew, p1ret, p2ret


def compress_profile_n(z, ps, delta_maxs):
    """Compresses multiple profiles, each profile: ps[i],
    by merging the neighbouring layers that
    has an density difference of maximum delta_maxs[i]"""
    inew, znew = compress_profile_index_n(z, ps, delta_maxs)
    psnew = create_compressed_profile(ps, inew)
    return znew, psnew


def compare_layers(ps, delta_max):
    """Compare consecutive layers and see if ps is smaller than delta_max"""
    if len(ps.shape) > 1:
        return (abs(ps[:-2:2] - ps[2::2]) > delta_max).any(axis=1)
    else:
        return abs(ps[:-2:2] - ps[2::2]) > delta_max


def compress_profile_index_n(z, ps, delta_maxs):
    znew = z.copy()
    inew = arange(len(ps[0]))
    psnew = [ps_i.copy() for ps_i in ps]
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        i += 1
        index = array([True] * len(psnew[0]))
        test = array([compare_layers(ps_i, delta_maxs_i) for ps_i, delta_maxs_i in zip(psnew, delta_maxs)])
        index[1:-1:2] = test.any(axis=0)

        psnew = [ps_i[index] for ps_i in psnew]
        znew = znew[index]
        inew = inew[index]
    inew = append(inew, inew[-1])
    return inew, znew


def create_compressed_profile(ps, inew):
    psret = [array([ps_i[inew[i] : inew[i + 1] + 1].mean(0) for i in range(len(inew) - 1)]) for ps_i in ps]
    return psret


def create_profile(d, sigma, dens, prof_funcs, dz=0.01, mult=3, buffer=0, delta=20):
    zlay = r_[0, cumsum(d)]
    z = arange(-sigma[0] * mult - buffer, zlay[-1] + sigma[-1] * mult + buffer, dz)
    ptot = z * 0
    pdens = z * 0
    ps = []
    d = r_[delta + sigma[0] * mult + buffer, d, sigma[-1] * mult + delta + buffer]
    sigma = r_[0, sigma, 0]
    zlay = cumsum(d)
    # print zlay
    z0 = buffer + sigma[1] * mult
    zlay = r_[0, zlay] - z0
    # print zlay
    # for i in range(len(d)):
    #    print z0
    #    p = make_profile(z, z0, d[i], sigma[i], sigma[i+1])
    #    z0 += d[i]
    #    ptot += p
    #    pdens += p*dens[i]
    #    ps.append(p)
    ps = array(
        [prof(z, zp, di, s_low, s_up) for prof, zp, di, s_low, s_up in zip(prof_funcs, zlay, d, sigma[:-1], sigma[1:])]
    )
    ptot = ps.sum(0)
    pdens_indiv = (ps * dens[:, newaxis]) / ptot
    pdens = pdens_indiv.sum(0)

    return z, pdens, pdens_indiv


def create_profile_cm(
    d,
    sigma_c,
    sigma_m,
    prof_funcs,
    prof_funcs_mag,
    dmag_dens_l,
    dmag_dens_u,
    mag_dens,
    dd_m,
    dz=0.01,
    mult=3,
    buffer=0,
    delta=20,
):
    """Create a scattering length profile for magnetic x-ray scattering for the charge and
    magnetic part of the scattering lengths. sigma is roughnesses.
    Note that prof_funcs relate to layer profiles and prof_funcs_mag relates to
    magnetic interface profiles.
    """
    zlay = r_[0, cumsum(d)]
    s_max_bot = max(sigma_c[0], sigma_m[0])
    s_max_up = max(sigma_c[-1], sigma_m[-1])
    z = arange(-s_max_bot * mult - buffer, zlay[-1] + s_max_up * mult + buffer + dz, dz)
    ptot = z * 0
    pdens = z * 0
    ps = []
    d = r_[delta + s_max_bot * mult + buffer, d, s_max_up * mult + delta + buffer]
    sigma_c = r_[0, sigma_c, 0] + 1e-20
    sigma_m = r_[0, sigma_m, 0] + 1e-20
    dd_m = r_[0, dd_m, 0] + 1e-20
    prof_funcs_mag = [prof_funcs_mag[0]] + prof_funcs_mag + [prof_funcs_mag[-1]]
    zlay = cumsum(d)
    z0 = delta + buffer + s_max_bot * mult
    zlay = r_[0, zlay] - z0
    ps = array(
        [
            r_[
                prof(z, zp, di, s_c_low, s_c_up),
                # prof(z, zp, di, s_m_low, s_m_up)]
                prof_m_l(-(z - zp - dd_low), s_m_low) * dm_l + prof_m_u(z - zp - di - dd_up, s_m_up) * dm_u + m,
            ]
            for prof, zp, di, s_c_low, s_c_up, s_m_low, s_m_up, prof_m_l, prof_m_u, dm_l, dm_u, m, dd_low, dd_up in zip(
                prof_funcs,
                zlay,
                d,
                sigma_c[:-1],
                sigma_c[1:],
                sigma_m[:-1],
                sigma_m[1:],
                prof_funcs_mag[:-1],
                prof_funcs_mag[1:],
                dmag_dens_l,
                dmag_dens_u,
                mag_dens,
                dd_m[:-1],
                dd_m[1:],
            )
        ]
    )
    ptot = ps.sum(0)
    # print ps.shape, ptot.shape
    pdens_indiv = (ps[:, : len(z)]) / ptot[: len(z)]
    pdens_indiv_c = pdens_indiv  # *dens_c[:,newaxis]
    pdens_indiv_m = ps[:, len(z) :]  # *dens_m[:,newaxis]
    # pdens_c = pdens_indiv_c.sum(0)
    # pdens_m = (pdens_indiv_m*pdens_indiv).sum(0)

    return z, pdens_indiv_c, pdens_indiv_m


def create_profile_cm2(
    d,
    sigma_c,
    sigma_ml,
    sigma_mu,
    prof_funcs,
    prof_funcs_mag,
    dmag_dens_l,
    dmag_dens_u,
    mag_dens,
    dd_l,
    dd_u,
    dz=0.01,
    mult=3,
    buffer=0,
    delta=20,
):
    """Create a scattering length profile for magnetic x-ray scattering for the charge and
    magnetic part of the scattering lengths. sigma is roughnesses.
    Note that prof_funcs relate to layer profiles and prof_funcs_mag relates to
    magnetic interface profiles.
    """
    zlay = r_[0, cumsum(d)]
    s_max_bot = max(sigma_c[0], sigma_ml[0])
    s_max_up = max(sigma_c[-1], sigma_mu[-1])
    z = arange(-s_max_bot * mult - buffer, zlay[-1] + s_max_up * mult + buffer + dz, dz)
    ptot = z * 0
    pdens = z * 0
    ps = []
    d = r_[delta + s_max_bot * mult + buffer, d, s_max_up * mult + delta + buffer]
    sigma_c = r_[0, sigma_c, 0] + 1e-20
    sigma_mu = r_[sigma_mu, 0] + 1e-20
    sigma_ml = r_[0, sigma_ml] + 1e-20

    dd_l = r_[dd_l, 0] + 1e-20
    dd_u = r_[dd_u, 0] + 1e-20

    prof_funcs_mag = [prof_funcs_mag[0]] + prof_funcs_mag + [prof_funcs_mag[-1]]
    zlay = cumsum(d)
    z0 = delta + buffer + s_max_bot * mult
    zlay = r_[0, zlay] - z0
    ps = array(
        [
            r_[
                prof(z, zp, di, s_c_low, s_c_up),
                # prof(z, zp, di, s_m_low, s_m_up)]
                prof_m_l(-(z - zp - dd_low), s_m_low) * dm_l * m
                + prof_m_u((z - zp - di + dd_up), s_m_up) * dm_u * m
                + m,
            ]
            for prof, zp, di, s_c_low, s_c_up, s_m_low, s_m_up, prof_m_l, prof_m_u, dm_l, dm_u, m, dd_low, dd_up in zip(
                prof_funcs,
                zlay,
                d,
                sigma_c[:-1],
                sigma_c[1:],
                sigma_ml,
                sigma_mu,
                prof_funcs_mag[:-1],
                prof_funcs_mag[1:],
                dmag_dens_l,
                dmag_dens_u,
                mag_dens,
                dd_l,
                dd_u,
            )
        ]
    )
    ptot = ps.sum(0)
    # print ps.shape, ptot.shape
    pdens_indiv = (ps[:, : len(z)]) / ptot[: len(z)]
    pdens_indiv_c = pdens_indiv  # *dens_c[:,newaxis]
    pdens_indiv_m = ps[:, len(z) :]  # *dens_m[:,newaxis]
    # pdens_c = pdens_indiv_c.sum(0)
    # pdens_m = (pdens_indiv_m*pdens_indiv).sum(0)

    return z, pdens_indiv_c, pdens_indiv_m


# if __name__ == "__main__":
#     from mpl_toolkits.axes_grid import Grid
#     from mpl_toolkits.axes_grid.anchored_artists import AnchoredText
#     from pylab import *
#
#     # sigma0 = 5.
#     # sigma1 = 1.
#     # mult = 3
#     # buffer = 0
#     # delta = 20
#     # z0 = sigma0*mult
#     # d = 9.#sigma0*mult + sigma1*mult
#     # z = arange(0, z0 + d + sigma1*mult, 0.1)
#     # p = make_profile(z, z0, d, sigma0, sigma1)
#     # plot(z,p)
#     sigma_c = array([4, 4, 2]) * 1.0  # sigma_c = array([2,8,2])*1.0
#     sigma_m = array([1, 1, 1]) * 1.0  # sigma_m = array([2,5,4])*1.0
#     dmag_l = array([0.0, 0.0, 0.0])
#     dmag_u = array([0.0, 0.0, 0.0])
#     mag_dens = array([0.0, 2.25, 0.0])
#     # ddm = array([10.0, -10.0, 0.0])
#     ddm = array([10.0, -10.0, 10.0])
#     d = array([0.0, 50, 0.0]) * 1.0  # d = array([5, 60])*1.0
#     dens_c = array([-1.79e-3 + 0.0029j, -4e-23])  # dens_c = array([5,50,25,0])
#     dens_m = array([1.4458e-3 - 0.00144j, 2.12e-23])  # dens_m = array([0,6,10,0])
#     prof_funcs = [erf_profile] * 3  # prof_funcs = [erf_profile]*4
#     mag_prof_funcs = [erf_interf] * 3
#     # z, pdens, pdens_indiv = create_profile(d, sigma_c, dens_cprof_funcs, dz = 0.1,
#     #                                       mult = 5.0,
#     #                                       buffer = 0, delta = 5.)
#     z, c_den, m_den = create_profile_cm(
#         d[1:-1],
#         sigma_c[:-1],
#         sigma_m[:-1],
#         prof_funcs,
#         mag_prof_funcs,
#         dmag_l,
#         dmag_u,
#         mag_dens,
#         ddm,
#         dz=1.0,
#         mult=4.0,
#         buffer=20,
#         delta=5.0,
#     )
#
#     def plot_fig(filename, fig_title=""):
#         fig = figure(1)
#         grid = Grid(fig, 111, nrows_ncols=(3, 1), axes_pad=0.15)
#         # subplot(311)
#         grid[0].set_title(fig_title)
#         grid[0].plot(z, c_den[0], lw=2.0, c="b")
#         grid[0].plot(z, c_den[1], lw=2.0, c="r")
#         grid[0].plot(z, c_den[2], lw=2.0, c="g")
#         grid[0].axvline(0, linestyle="--", c="k")
#         grid[0].axvline(d[1], linestyle="--", c="k")
#         grid[0].set_ylabel("Comp.")
#         grid[0].legend(("Substrate", "Fe layer", "Ambient"))
#         # subplot(312)
#         grid[1].plot(z, m_den[1], lw=2.0, c="r")
#         grid[1].axvline(0, linestyle="--", c="k")
#         grid[1].axvline(d[1], linestyle="--", c="k")
#         grid[1].axvline(0 + ddm[0], linestyle=":", c="k")
#         grid[1].axvline(d[1] + ddm[1], linestyle=":", c="k")
#         # grid[1].annotate('Sub.dd_m = %.1f'%ddm[0], xy=(ddm[0], mag_dens[1]), xycoords = 'data',
#         #                 xytext = (-50, 0), textcoords = 'offset points',
#         #                 arrowprops=dict(arrowstyle="->", connectionstyle = 'arc3'))
#         # grid[1].annotate('Fe.dd_m = %.1f'%ddm[1], xy=(d[1]+ddm[1], mag_dens[1]), xycoords = 'data',
#         #                 xytext = (50, 0), textcoords = 'offset points',
#         #                 arrowprops=dict(arrowstyle="<-", connectionstyle = 'arc3'))
#         at = AnchoredText(
#             "dd_ml = %.1f" % (ddm[0]),
#             prop=dict(size=10),
#             frameon=True,
#             loc=6,
#         )
#         at.patch.set_boxstyle("round,pad=0.,rounding_size=0.2")
#         grid[1].add_artist(at)
#         at = AnchoredText(
#             "dd_mu = %.1f" % (ddm[1]),
#             prop=dict(size=10),
#             frameon=True,
#             loc=7,
#         )
#         at.patch.set_boxstyle("round,pad=0.,rounding_size=0.2")
#         grid[1].add_artist(at)
#         grid[1].set_ylabel("Mag. mom.\n per res. atom.")
#         grid[1].set_ylim(0, 3.1)
#         # subplot(313)
#         grid[2].plot(z, m_den[1] * c_den[1], "r", z, 2.25 * c_den[1], "b", lw=2.0)
#         grid[2].axvline(0, linestyle="--", c="k")
#         grid[2].axvline(d[1], linestyle="--", c="k")
#         grid[2].axvline(0 + ddm[0], linestyle=":", c="k")
#         grid[2].axvline(d[1] + ddm[1], linestyle=":", c="k")
#         grid[2].set_ylabel("Tot. Mag mom.")
#         grid[2].legend(("Profile", "Bulk Fe"))
#         grid[2].set_xlabel("Depth [AA]")
#         savefig(filename)
#         clf()
#
#     plot_fig("Bulk_Fe_sigma", "Bulk Fe with no change at the interfaces")
#     dmag_l = array([0.0, 0.75, 0.0])
#     dmag_u = array([0.0, 0.75, 0.0])
#     z, c_den, m_den = create_profile_cm(
#         d[1:-1],
#         sigma_c[:-1],
#         sigma_m[:-1],
#         prof_funcs,
#         mag_prof_funcs,
#         dmag_l,
#         dmag_u,
#         mag_dens,
#         ddm,
#         dz=1.0,
#         mult=4.0,
#         buffer=20,
#         delta=5.0,
#     )
#     plot_fig("Enhanced_Fe_sigma", "Bulk Fe with enhanced interfaces")
#     dmag_l = array([0.0, -1.0, 0.0])
#     dmag_u = array([0.0, -1.0, 0.0])
#     z, c_den, m_den = create_profile_cm(
#         d[1:-1],
#         sigma_c[:-1],
#         sigma_m[:-1],
#         prof_funcs,
#         mag_prof_funcs,
#         dmag_l,
#         dmag_u,
#         mag_dens,
#         ddm,
#         dz=1.0,
#         mult=4.0,
#         buffer=20,
#         delta=5.0,
#     )
#     plot_fig("Depleted_Fe_sigma", "Bulk Fe with depleted interfaces")
